from __future__ import annotations

import asyncio
import functools
import inspect
import multiprocessing

import os
import random
import signal
import socket
import sys
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from inspect import signature
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_args,
    get_type_hints,
    get_origin,
)


import msgspec

from mkfst.connection.tcp.fabricator import Fabricator
from mkfst.connection.tcp.mercury_sync_http_connection import (
    MercurySyncHTTPConnection,
)
from mkfst.docs import (
    ParsedAPIMetadata,
    ParsedAPIServer,
    ParsedEndpoint,
    ParsedEndpointMetadata,
    ParsedOperationMetadata,
    ParsedServerVariable,
    ParsedTag,
    create_api_definition,
    get_redoc_html,
    get_swagger_ui_html,
)
from mkfst.env import Env, load_env
from mkfst.hooks import endpoint
from mkfst.logging import Logger, LogLevelName, LoggingConfig
from mkfst.middleware.base import Middleware
from mkfst.middleware.base.base_wrapper import BaseWrapper
from mkfst.models.http import HTML, Model
from mkfst.models.logging import Event

from mkfst.tasks import TaskRunner
from .group import Group
from .parse_to_source_type import parse_to_source_type
from .socket import bind_tcp_socket


Json = dict | list


multiprocessing.allow_connection_pickling()
spawn = multiprocessing.get_context("spawn")


E = TypeVar("E", bound=Env)

ServerVariable = Dict[Literal["options", "default", "description"], List[str] | str]

Tag = Dict[Literal["value", "description", "docs_description", "docs_url"], str]


async def run(
    service_name: str,
    tcp_connection: MercurySyncHTTPConnection,
    env: msgspec.Struct,
    groups: List[Group],
    config: Dict[str, Union[int, socket.socket, str]] = {},
):
    try:
        tcp_connection.from_env(env)

        services = {cls.__name__: cls for cls in Service.__subclasses__()}
        service: Type[Service] = services.get(service_name)

        server = service(
            tcp_connection.host,
            tcp_connection.port,
            upgrade_port=config.get("upgrade_port"),
            cert_path=config.get("cert_path"),
            key_path=config.get("key_path"),
            groups=groups,
            env=env,
            log_level=config.get("log_level", "info"),
            middleware=config.get("middleware"),
        )

        await server.run(
            worker_socket=config.get("tcp_socket"),
            upgrade_socket=config.get("upgrade_socket"),
        )

    except Exception:
        pass

    current_task = asyncio.current_task()

    tasks = asyncio.all_tasks()
    for task in tasks:
        if task != current_task:
            try:
                task.cancel()

            except (
                Exception,
                asyncio.InvalidStateError,
                asyncio.CancelledError,
                asyncio.TimeoutError,
                AssertionError,
            ):
                pass

    try:
        await asyncio.gather(
            *[task for task in tasks if task != current_task], return_exceptions=True
        )

    except Exception:
        pass


def start_pool(
    tcp_connection: MercurySyncHTTPConnection,
    service_name: str,
    custom_env: msgspec.Struct,
    groups: List[Group],
    config: Dict[str, Union[int, socket.socket, str]] = {},
):
    import asyncio

    try:
        import uvloop

        uvloop.install()

    except ImportError:
        pass

    try:
        loop = asyncio.get_event_loop()

    except Exception:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    stdin_fileno = config.get("stdin_fileno")

    if stdin_fileno is not None:
        sys.stdin = os.fdopen(os.dup(stdin_fileno))

    try:
        loop = asyncio.get_event_loop()
        loop.run_until_complete(
            run(
                service_name,
                tcp_connection,
                custom_env,
                groups,
                config,
            )
        )

    except Exception:
        pass


class Service(Generic[E]):
    def __init__(
        self,
        host: str,
        port: int,
        upgrade_port: int | None = None,
        cert_path: Optional[str] = None,
        key_path: Optional[str] = None,
        workers: int = os.cpu_count(),
        env: Optional[E] = None,
        engine: Literal["process", "async"] = "process",
        groups: List[Group] | None = None,
        log_level: LogLevelName = "info",
        middleware: List[Middleware] | None = None,
        service_metadata: Dict[
            Literal[
                "name",
                "version",
                "summary",
                "owner",
                "owner_url",
                "owner_email",
                "license",
                "license_identifier",
                "license_url",
                "server_url",
                "server_description",
                "server_variables",
            ],
            str | Dict[str, ServerVariable],
        ]
        | None = None,
        tags: List[Tag] | None = None,
    ) -> None:
        env = load_env(Env, override=env)

        self.env = env
        self.name = self.__class__.__name__
        self._logging_config = LoggingConfig(level=log_level)
        self.logger = Logger()
        self.logger.configure(
            name=self.name,
            level=log_level,
        )

        self._instance_id = random.randint(0, 2**16)

        if service_metadata is None:
            service_metadata = {}

        self._service_metadata = service_metadata
        self._tags = tags or []

        self._service_name = self._service_metadata.get("name", self.name)

        self._tags.append(
            ParsedTag(
                value=self._service_name,
                description=f"The {self._service_name} service group.",
            )
        )

        if groups:
            for group in groups:
                self._tags.extend(group.tags)

        if workers < 1:
            workers = 1

        self._workers = workers

        self.host = host
        self.port = port

        self.upgrade_port: int | None = None
        if upgrade_port and upgrade_port != port:
            self.upgrade_port = upgrade_port

        self.cert_path = cert_path
        self.key_path = key_path

        if groups is None:
            groups = []

        if middleware is None:
            middleware = []

        self.middleware = middleware
        self.groups = groups

        self._is_worker = False
        self._engine: Union[ProcessPoolExecutor, None] = None
        self._tcp_queue: Dict[Tuple[str, int], asyncio.Queue] = defaultdict(
            asyncio.Queue
        )
        self._cleanup_task: Union[asyncio.Task, None] = None
        self._waiter: Union[asyncio.Future, None] = None

        self.engine_type = engine

        self.instance_ids = [self._instance_id + idx for idx in range(0, workers)]
        self._tcp = MercurySyncHTTPConnection(
            self.host,
            self.port,
            self._instance_id,
            upgrade_port=upgrade_port,
            env=self.env,
        )

        self.tasks = TaskRunner(self._instance_id, env)

        self._handlers: Dict[str, Callable[..., Awaitable[Any]]] = {}
        self._response_headers: Dict[str, Dict[str, Any]] = {}

        self._endpoint_docs: Dict[str, ParsedEndpoint] = {}
        self._docs_json: str | None = None
        self._group_middleware: List[Middleware] = []
        self._reserved_urls = [
            "/openapi.json",
            "/docs",
            "/docs/oauth2-redirect",
            "/redoc",
        ]

        self._loop: asyncio.AbstractEventLoop | None = None

        self._setup()
        self._apply_groups()
        self._create_docs()

    @endpoint("/api/status")
    async def status(self) -> Literal["OK"]:
        return "OK"

    @endpoint("/openapi.json")
    async def get_openapi_json(self) -> dict:
        return self._docs_json

    @endpoint("/docs")
    async def get_docs(self) -> HTML:
        title = self._service_metadata.get("name", self.name)

        return get_swagger_ui_html(
            openapi_url="/openapi.json", title=f"{title} - Swagger UI"
        )

    @endpoint("/redoc")
    async def get_redocs(self) -> HTML:
        title = self._service_metadata.get("name", self.name)

        return get_redoc_html(
            openapi_url="/openapi.json", title=f"{title} - Swagger UI"
        )

    @endpoint("/docs/oauth2-redirect", response_headers={"Content-Type": "text/html"})
    async def get_docs_auth_redirect(self) -> HTML:
        return

    def _create_docs(self):
        server_url = self._service_metadata.get("server_url")
        if server_url is None and self.env.MERCURY_SYNC_SERVER_URL:
            server_url = self.env.MERCURY_SYNC_SERVER_URL.unicode_string()

        elif self.cert_path and self.key_path:
            server_url = f"https://{self.host}:{self.port}"

        else:
            server_url = f"http://{self.host}:{self.port}"

        server_variables: Dict[str, ParsedServerVariable] | None = None
        variables_config: Dict[str, ServerVariable] = self._service_metadata.get(
            "server_variables"
        )
        if isinstance(variables_config, dict) and len(variables_config) > 0:
            server_variables = {
                variable_name: ParsedServerVariable(
                    options=variable_config.get("options"),
                    default=variable_config.get("default"),
                    description=variable_config.get("description"),
                )
                for variable_name, variable_config in variables_config.items()
            }

        self._docs_json = create_api_definition(
            ParsedAPIMetadata(
                title=self._service_metadata.get("name", self.name),
                version=self._service_metadata.get(
                    "version", self.env.MERCURY_SYNC_API_VERISON
                ),
                summary=self._service_metadata.get("summary"),
                description=self.__doc__,
                owner=self._service_metadata.get("owner"),
                owner_url=self._service_metadata.get("owner_url"),
                owner_email=self._service_metadata.get("owner_email"),
                license=self._service_metadata.get("license"),
                license_identifier=self._service_metadata.get("license_identifier"),
                license_url=self._service_metadata.get("license_url"),
            ),
            ParsedAPIServer(
                server_url=server_url,
                server_description=self._service_metadata.get("server_description"),
                server_variables=server_variables,
            ),
            self._endpoint_docs,
            tags=self._tags,
        )

    def _apply_groups(self):
        for group in self.groups:
            assembled: Dict[
                Literal[
                    "routes",
                    "match_routes",
                    "events",
                    "parsers",
                    "middleware_enabled",
                    "supported_handlers",
                    "response_parsers",
                    "request_parsers",
                    "fabricators",
                    "endpoint_docs",
                    "middleware",
                    "response_headers",
                ],
                Dict[str, Any],
            ] = group._assemble(
                self._instance_id,
                self.env,
                self.middleware,
            )

            self._apply_group(
                assembled,
            )

    def _apply_group(
        self,
        assembled: Dict[
            Literal[
                "routes",
                "match_routes",
                "events",
                "parsers",
                "middleware_enabled",
                "supported_handlers",
                "response_parsers",
                "request_parsers",
                "fabricators",
                "endpoint_docs",
                "middleware",
                "response_headers",
            ],
            Dict[str, Any],
        ],
    ):
        events = assembled["events"]
        fabricators = assembled["fabricators"]
        request_parsers = assembled["request_parsers"]
        response_parsers = assembled["response_parsers"]
        supported_handlers = assembled["supported_handlers"]
        middleware_enabled = assembled["middleware_enabled"]
        response_headers = assembled["response_headers"]
        endpoint_docs = assembled["endpoint_docs"]

        self._tcp.fabricators.update(
            {
                f"{method}_{route}": fabricator
                for route, methods in fabricators.items()
                for method, fabricator in methods.items()
            }
        )

        self._tcp._request_parsers.update(
            {
                f"{method}_{route}": parser
                for route, methods in request_parsers.items()
                for method, parser in methods.items()
            }
        )

        self._tcp._response_parsers.update(
            {
                f"{method}_{route}": parser
                for route, methods in response_parsers.items()
                for method, parser in methods.items()
            }
        )

        self._tcp._supported_handlers.update(
            {
                f"{method}_{route}": supported_handler
                for route, methods in supported_handlers.items()
                for method, supported_handler in methods.items()
            }
        )

        self._tcp._middleware_enabled.update(
            {
                f"{method}_{route}": middleware_enabled_status
                for route, methods in middleware_enabled.items()
                for method, middleware_enabled_status in methods.items()
            }
        )

        self._tcp._response_headers.update(
            {
                f"{method}_{route}": response_header_set
                for route, methods in response_headers.items()
                for method, response_header_set in methods.items()
            }
        )

        self._endpoint_docs.update(
            {route: endpoint_doc for route, endpoint_doc in endpoint_docs.items()}
        )

        for route, methods in assembled["routes"].items():
            self._tcp.events.update(
                {f"{method}_{route}": events[route] for method in methods}
            )

            self._tcp.routes.add(route, methods)

        self._group_middleware.extend(
            [
                middleware
                for middleware in assembled["middleware"]
                if middleware not in self._group_middleware
            ]
        )

    def _setup(self):
        response_parsers: Dict[Model, Tuple[Callable[[Any], str], int]] = {}
        request_parsers: Dict[str, Model] = {}

        routes = {}

        (
            endpoints,
            tasks,
            fabricators,
        ) = self._gather_hooks()

        for task in tasks.values():
            self.tasks.add(task)

        for path, path_endpoint in endpoints.items():
            handler = path_endpoint
            has_middleware = len(self.middleware) > 0
            if has_middleware and handler.path not in self._reserved_urls:
                self._tcp._middleware_enabled[path] = True

            if isinstance(path_endpoint, (Middleware, BaseWrapper)):
                handler = self._handlers.get(path)

            endpoint_signature = signature(handler)
            params = endpoint_signature.parameters.values()

            return_type = get_type_hints(handler).get("return")
            response_types = get_args(return_type)

            request_parsers.update(
                {
                    path: get_args(param_type.annotation)[0]
                    for param_type in params
                    if (
                        (args := get_args(param_type.annotation))
                        and len(args) > 0
                        and args[0] in Model.__subclasses__()
                    )
                }
            )

            routes[handler.path] = {method: path_endpoint for method in handler.methods}
            methods = handler.methods

            if (
                len(response_types) > 1
                and inspect.isclass(response_types[0])
                and response_types[0] in Model.__subclasses__()
            ):
                model = response_types[0]
                status_code = response_types[1]

                response_parsers.update(
                    {
                        f"{method}_{handler.path}": (model, status_code)
                        for method in methods
                    }
                )

            elif (
                len(response_types) > 0 and response_types[0] in Model.__subclasses__()
            ):
                model = response_types[0]

                response_parsers.update(
                    {f"{method}_{handler.path}": (model, 200) for method in methods}
                )

            elif return_type in Model.__subclasses__():
                response_parsers.update(
                    {
                        f"{method}_{handler.path}": (return_type, 200)
                        for method in methods
                    }
                )

            elif return_type in get_args(Json) or get_origin(return_type) in get_args(
                Json
            ):
                response_parsers.update(
                    {
                        f"{method}_{handler.path}": (
                            parse_to_source_type(
                                return_type,
                            ),
                            200,
                        )
                        for method in methods
                    }
                )

            if isinstance(handler.responses, dict):
                responses = handler.responses

                response_parsers.update(
                    {
                        f"{method}_{handler.path}": (response_model, status)
                        for method in methods
                        for status, response_model in responses.items()
                        if (issubclass(response_model, Model))
                    }
                )

        self._tcp.events.update(endpoints)
        self._tcp.match_routes.update(routes)
        self._tcp.fabricators.update(fabricators)
        self._tcp._response_headers.update(self._response_headers)

        for route, methods in routes.items():
            self._tcp.routes.add(route, methods)

        for key, parser in request_parsers.items():
            self._tcp._request_parsers[key] = parser

        for key, parser in response_parsers.items():
            self._tcp._response_parsers[key] = parser

    def _gather_hooks(self):
        reserved = ["connect", "close"]

        endpoints: Dict[
            str, Callable[..., Awaitable[Model | Dict[Any, Any] | str]]
        ] = {}
        fabricators: Dict[str, Fabricator] = {}
        tasks: Dict[str, Callable[[], Awaitable[Any]]] = {}

        for _, call in inspect.getmembers(self, predicate=inspect.ismethod):
            hook_name: str = call.__name__
            not_internal = hook_name.startswith("__") is False
            not_reserved = hook_name not in reserved
            is_endpoint = hasattr(call, "as_endpoint")
            is_task = hasattr(call, "as_task")

            if not_internal and not_reserved and is_endpoint:
                methods: List[str] = call.methods
                path: str = call.path

                handler = call
                for middleware_operator in self.middleware:
                    if path not in self._reserved_urls:
                        call = middleware_operator.wrap(call)

                endpoints.update({f"{method}_{path}": call for method in methods})

                self._handlers.update(
                    {f"{method}_{path}": handler for method in methods}
                )

                call_params = inspect.signature(handler).parameters

                required_params = [
                    (key, value.annotation)
                    for key, value in call_params.items()
                    if value.default == inspect._empty
                ]

                optional_params = {
                    key: get_args(value.annotation)
                    for key, value in call_params.items()
                    if value.default != inspect._empty
                }

                fabricator = Fabricator(required_params, optional_params)

                return_type = get_type_hints(handler).get("return")
                response_types = get_args(return_type)

                responses: Dict[int, Any] = {}

                if (
                    len(response_types) > 1
                    and inspect.isclass(response_types[0])
                    and response_types[0] in Model.__subclasses__()
                ):
                    model = response_types[0]
                    status_code = response_types[1]

                    responses[status_code] = model

                elif len(response_types) == 1:
                    responses[200] = response_types[0]

                elif (
                    len(response_types) > 0
                    and response_types[0] in Model.__subclasses__()
                ):
                    model = response_types[0]
                    responses[200] = model

                else:
                    responses[200] = return_type

                if isinstance(handler.responses, dict):
                    responses.update(handler.responses)

                additional_docs: Dict[Literal["docs_description", "docs_url"], str] = {}

                if isinstance(handler.additional_docs, dict):
                    additional_docs = handler.additional_docs

                method_metadata: Dict[
                    Literal[
                        "GET",
                        "HEAD",
                        "OPTIONS",
                        "POST",
                        "PUT",
                        "PATCH",
                        "DELETE",
                        "TRACE",
                    ],
                    Dict[
                        Literal[
                            "name",
                            "tags",
                            "description",
                            "summary",
                            "docs_url",
                            "docs_description",
                            "depreciated",
                        ],
                        str | List[str] | bool,
                    ],
                ] = handler.method_metadata

                call_response_headers = dict(handler.response_headers)

                self._response_headers.update(
                    {
                        f"{method}_{path}": dict(call_response_headers)
                        for method in methods
                    }
                )

                if path not in self._reserved_urls:
                    self._endpoint_docs[path] = ParsedEndpoint(
                        path=path,
                        methods=methods,
                        endpoint_metadata=ParsedEndpointMetadata(
                            description=handler.__doc__,
                            summary=handler.summary,
                            operations={
                                method: ParsedOperationMetadata(
                                    group_name=self._service_name,
                                    name=handler.__name__,
                                    tags=method_metadata[method].get("tags"),
                                    description=method_metadata[method].get(
                                        "description"
                                    ),
                                    docs_description=additional_docs.get(
                                        "docs_description"
                                    ),
                                    docs_url=additional_docs.get("docs_url"),
                                    deprecated=method_metadata[method].get(
                                        "depreciated"
                                    ),
                                )
                                for method in methods
                            },
                        ),
                        responses={
                            status: parse_to_source_type(response_type)
                            for status, response_type in responses.items()
                        },
                        response_headers=call_response_headers,
                        headers=parse_to_source_type(fabricator["headers"]),
                        parameters=parse_to_source_type(fabricator["parameters"]),
                        query=parse_to_source_type(fabricator["query"]),
                        body=parse_to_source_type(fabricator["body"]),
                        required=fabricator.required_params,
                    )

                for method in methods:
                    endpoint_method_key = f"{method}_{path}"
                    fabricators[endpoint_method_key] = fabricator

                self._tcp._supported_handlers[path] = {
                    method: handler for method in methods
                }

            if not_internal and not_reserved and is_task:
                hook_name = handler.__name__
                tasks[hook_name] = handler

        return (endpoints, tasks, fabricators)

    async def run(
        self,
        upgrade_socket: Optional[socket.socket] = None,
        worker_socket: Optional[socket.socket] = None,
    ):
        self._loop = asyncio.get_event_loop()
        for signame in ("SIGINT", "SIGTERM", "SIG_IGN"):
            self._loop.add_signal_handler(
                getattr(
                    signal,
                    signame,
                ),
                self.abort,
            )

        self._is_worker = worker_socket is not None

        async with self.logger.context(
            name=self.name,
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            if worker_socket is None:
                await ctx.log(Event(message="Startiworker_socketng server"))

            await asyncio.gather(
                *[middleware.__setup__() for middleware in self.middleware]
            )

            await asyncio.gather(
                *[middleware.__setup__() for middleware in self._group_middleware]
            )

            pool: List[asyncio.Future] = []

            loop = asyncio.get_event_loop()

            try:
                stdin_fileno = sys.stdin.fileno()
            # The `sys.stdin` can be `None`, see https://docs.python.org/3/library/sys.html#sys.__stdin__.
            except (AttributeError, OSError):
                stdin_fileno = None

            if (
                self.engine_type == "process"
                and worker_socket is None
                and self._workers > 1
            ):
                await ctx.log(
                    Event(message=f"Initializing - {self._workers} - workers")
                )

                engine = ProcessPoolExecutor(
                    max_workers=self._workers,
                    mp_context=spawn,
                )

                tcp_socket = bind_tcp_socket(self.host, self.port)
                upgrade_socket: socket.socket | None = None
                if self.upgrade_port and upgrade_socket is None:
                    upgrade_socket = bind_tcp_socket(self.host, self.upgrade_port)

                config = {
                    "tcp_socket": tcp_socket,
                    "upgrade_socket": upgrade_socket,
                    "stdin_fileno": stdin_fileno,
                    "cert_path": self.cert_path,
                    "key_path": self.key_path,
                    "log_level": self._logging_config.level.name.lower(),
                    "middleware": self.middleware,
                    "upgrade_port": self.upgrade_port,
                }

                for _ in range(self._workers):
                    connection = MercurySyncHTTPConnection(
                        self.host,
                        self.port,
                        self._instance_id,
                        upgrade_port=self.upgrade_port,
                        env=self.env,
                    )

                    service_name = self.__class__.__name__

                    service_worker = loop.run_in_executor(
                        engine,
                        functools.partial(
                            start_pool,
                            connection,
                            service_name,
                            self.env,
                            self.groups,
                            config=config,
                        ),
                    )

                    pool.append(service_worker)

                try:
                    await asyncio.gather(*pool)

                except (
                    asyncio.CancelledError,
                    BrokenPipeError,
                    OSError,
                    KeyboardInterrupt,
                    Exception,
                ):
                    pass

                await self.close()

            else:
                if self.upgrade_port and upgrade_socket is None:
                    upgrade_socket = bind_tcp_socket(self.host, self.upgrade_port)

                await self._tcp.connect_async(
                    cert_path=self.cert_path,
                    key_path=self.key_path,
                    worker_socket=worker_socket,
                    upgrade_socket=upgrade_socket,
                )

                self.start_tasks()

                if worker_socket:
                    await ctx.log(
                        Event(message=f"Worker running on - {self.host}:{self.port}")
                    )

                self._waiter = self._loop.create_future()
                try:
                    if self._is_worker is False:
                        await self.logger.log(
                            Event(message=f"Main running on - {self.host}:{self.port}"),
                            template="{timestamp} - {level} - {thread_id} - {message}",
                        )

                    await self._waiter

                except Exception:
                    pass

                await self.close()

    def start_tasks(self):
        self.tasks.start_cleanup()
        for task in self.tasks.all_tasks():
            task.call = task.call.__get__(self, self.__class__)
            setattr(self, task.name, task.call)

            if task.trigger == "ON_START":
                self.tasks.start(task.name)

    async def close(self) -> None:
        async with self.logger.context(
            name=self.name, template="{timestamp} - {level} - {thread_id} - {message}"
        ) as ctx:
            if self._engine and not self._is_worker:
                await ctx.log(
                    Event(message=f"Shutting down - {self._workers} - workers")
                )
                self._engine.shutdown(cancel_futures=True)

            await ctx.log(
                Event(message=f"Shutting down tcp at - {self.host}:{self.port}")
            )
            await self._tcp.close()

            if self._waiter and not self._waiter.done():
                self._waiter.set_result(None)

        await asyncio.gather(*[middleware.close() for middleware in self.middleware])

        await asyncio.gather(
            *[middleware.close() for middleware in self._group_middleware]
        )

        if not self._is_worker:
            await ctx.log(Event(message="Graceful shutdown complete. Goodbye!"))

        await self.logger.close()

    def abort(self, *args):
        if self._engine:
            self._engine.shutdown(cancel_futures=True)

        self._tcp.abort()

        if self._waiter:
            self._waiter.set_result(None)

        self.logger.abort()

        for middleware in self.middleware:
            middleware.abort()

        for middleware in self._group_middleware:
            middleware.abort()



from __future__ import annotations
import collections
from functools import lru_cache
from typing import (
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Union,
)

# This is a nested structure similar to a linked-list
VariablePartsType = Tuple[tuple, Tuple[str, str]]


class RouteError(Exception):
    """Base error for any exception raised by Kua"""


def normalize_url(url: str) -> str:
    """
    Remove leading and trailing slashes from a URL

    :param url: URL
    :return: URL with no leading and trailing slashes

    :private:
    """
    if url.startswith("/"):
        url = url[1:]

    if url.endswith("/"):
        url = url[:-1]

    return url


def _unwrap(variable_parts: VariablePartsType):
    """
    Yield URL parts. The given parts are usually in reverse order.
    """
    curr_parts = variable_parts
    var_any = []

    while curr_parts:
        curr_parts, (var_type, part) = curr_parts

        if var_type == Router._VAR_ANY_NODE:
            var_any.append(part)
            continue

        if var_type == Router._VAR_ANY_BREAK:
            if var_any:
                yield tuple(reversed(var_any))
                var_any.clear()

            var_any.append(part)
            continue

        if var_any:
            yield tuple(reversed(var_any))
            var_any.clear()
            yield part
            continue

        yield part

    if var_any:
        yield tuple(reversed(var_any))


def make_params(
    key_parts: Sequence[str], variable_parts: VariablePartsType
) -> Dict[str, Union[str, Tuple[str]]]:
    """
    Map keys to variables. This map\
    URL-pattern variables to\
    a URL related parts

    :param key_parts: A list of URL parts
    :param variable_parts: A linked-list\
    (ala nested tuples) of URL parts
    :return: The param dict with the values\
    assigned to the keys

    :private:
    """
    # The unwrapped variable parts are in reverse order.
    # Instead of reversing those we reverse the key parts
    # and avoid the O(n) space required for reversing the vars
    return dict(zip(reversed(key_parts), _unwrap(variable_parts)))


class Route(NamedTuple):
    anything: Any
    key_parts: List[str]
    route: str


class ResolvedRoute(NamedTuple):
    anything: Any
    params: Dict[str, str | Tuple[str]]
    route: str


_Route = collections.namedtuple("_Route", ["route", "key_parts", "anything"])
RouteResolved = collections.namedtuple("RouteResolved", ["params", "anything", "route"])
RouteResolved.__doc__ = """
    Resolved route

    :param dict params: Pattern variables\
    to URL parts
    :param object anything: Literally anything.\
    This is attached to the URL pattern when\
    registering it
    """


class Router:
    """
    Route URLs to registered URL patterns.

    Thread safety: adding routes is not thread-safe,\
    it should be done on import time, everything else is.

    URL matcher supports ``:var`` for matching dynamic\
    path parts and ``:*var`` for matching one or more parts.

    Path parts are matched in the following order: ``static > var > any-var``.

    Usage::

        routes = kua.Routes()
        routes.add('api/:foo', {'GET': my_get_controller})
        route = routes.match('api/hello-world')
        route.params
        # {'foo': 'hello-world'}
        route.anything
        # {'GET': my_get_controller}

        # Matching any path
        routes.add('assets/:*foo', {})
        route = routes.match('assets/user/profile/avatar.jpg')
        route.params
        # {'foo': ('user', 'profile', 'avatar.jpg')}

        # Error handling
        try:
            route = routes.match('bad-url/some')
        except kua.RouteError:
            raise ValueError('Not found 404')
        else:
            # Do something useful here
            pass

    :ivar max_depth: The maximum URL depth\
    (number of parts) willing to match. This only\
    takes effect when one or more URLs matcher\
    make use of any-var (i.e: ``:*var``), otherwise the\
    depth of the deepest URL is taken.
    """

    _VAR_NODE = ":var"
    _VAR_ANY_NODE = ":*var"
    _ROUTE_NODE = ":route"

    _VAR_ANY_BREAK = ":*break"

    def __init__(
        self,
        max_depth: int = 40,
    ) -> None:
        """
        :ivar _routes: \
        Contain a graph with the parts of\
        each URL pattern. This is referred as\
        "partial route" later in the docs.
        :vartype _routes: dict
        :ivar _max_depth: Depth of the deepest\
        registered pattern
        :vartype _max_depth: int

        :private-vars:
        """
        # Routes graph format for 'foo/:foobar/bar':
        # {
        #   'foo': {
        #       ':var': {
        #           'bar': {
        #               ':route': _Route(),
        #               ...
        #           },
        #           ...
        #        }
        #        ...
        #   },
        #   ...
        # }
        self._routes: dict[str, RouteResolved] = {}
        self._max_depth = max_depth

    @lru_cache(maxsize=1024)
    def match(self, url: str) -> ResolvedRoute:
        """
        Match URL parts to a registered pattern.

        This function is basically where all\
        the CPU-heavy work is done.

        :param parts: URL parts
        :return: Matched route
        :raises kua.routes.RouteError: If there is no match

        :private:
        """

        if url.startswith("/"):
            url = url[1:]

        if url.endswith("/"):
            url = url[:-1]

        parts = url.split("/", self._max_depth + 1)

        if (len(parts) - 1) > self._max_depth:
            return None

        route_match: Optional[Route] = None  # type: RouteResolved
        route_variable_parts = tuple()  # type: VariablePartsType
        # (route_partial, variable_parts, depth)
        to_visit = [(self._routes, tuple(), 0)]  # type: List[Tuple[dict, tuple, int]]

        # Walk through the graph,
        # keep track of all possible
        # matching branches and do
        # backtracking if needed
        while to_visit:
            curr, curr_variable_parts, depth = to_visit.pop()
            try:
                part = parts[depth]
            except IndexError:
                if self._ROUTE_NODE in curr:
                    route_match = curr[self._ROUTE_NODE]
                    route_variable_parts = curr_variable_parts
                    break
                else:
                    continue

            if self._VAR_ANY_NODE in curr:
                to_visit.append(
                    (
                        {self._VAR_ANY_NODE: curr[self._VAR_ANY_NODE]},
                        (curr_variable_parts, (self._VAR_ANY_NODE, part)),
                        depth + 1,
                    )
                )
                to_visit.append(
                    (
                        curr[self._VAR_ANY_NODE],
                        (curr_variable_parts, (self._VAR_ANY_BREAK, part)),
                        depth + 1,
                    )
                )

            if self._VAR_NODE in curr:
                to_visit.append(
                    (
                        curr[self._VAR_NODE],
                        (curr_variable_parts, (self._VAR_NODE, part)),
                        depth + 1,
                    )
                )

            if part in curr:
                to_visit.append((curr[part], curr_variable_parts, depth + 1))

        if not route_match:
            return None

        return RouteResolved(
            route=route_match.route,
            params=make_params(
                key_parts=route_match.key_parts, variable_parts=route_variable_parts
            ),
            anything=route_match.anything,
        )

    def add(self, full_url: str, anything: Any) -> None:
        """
        Register a URL pattern into\
        the routes for later matching.

        It's possible to attach any kind of\
        object to the pattern for later\
        retrieving. A dict with methods and callbacks,\
        for example. Anything really.

        Registration order does not matter.\
        Adding a URL first or last makes no difference.

        :param url: URL
        :param anything: Literally anything.
        """
        url = normalize_url(full_url)
        parts = url.split("/")
        curr_partial_routes = self._routes
        curr_key_parts = []

        for part in parts:
            if part.startswith(":*"):
                curr_key_parts.append(part[2:])
                part = self._VAR_ANY_NODE

            elif part.startswith(":"):
                curr_key_parts.append(part[1:])
                part = self._VAR_NODE

            curr_partial_routes = curr_partial_routes.setdefault(part, {})

        curr_partial_routes[self._ROUTE_NODE] = _Route(
            route=full_url,
            key_parts=curr_key_parts,
            anything=anything,
        )

        self._max_depth = max(self._max_depth, len(parts) - 1)



import re
from typing import (
    Any,
    Dict,
    List,
    Literal,
    Set,
    get_args,
    get_origin,
)


from mkfst.models import (
    HTML,
    Body,
    FileUpload,
    Headers,
    Parameters,
    Query,
    Model,
)
from mkfst.models.http.request_models import HTTPEncodable
from .response_models import (
    InternalErrorSet,
    BadRequestErrorSet,
)

from .models import (
    Header,
    MediaType,
    Operation,
    Parameter,
    PathItem,
    RequestBody,
    Response,
)
from .parse_param import parse_param
from .parsed_endpoint_metadata import (
    HTTPMethod,
    ParsedEndpointMetadata,
)
from .parsed_operation_metadata import ParsedOperationMetadata
from .parsed_tag import ParsedTag
from .parsed_types import FieldMetadata, SchemaType, FieldType
from .validation_errors import ValidationErrorGroup


Json = dict | list
Raw = str | bytes


JSON = Dict[HTTPEncodable, HTTPEncodable] | List[HTTPEncodable]
REF_TEMPLATE = "#/components/schemas/{model}"


ParamType = type[Parameter] | type[Header] | type[Query]

path_pattern = re.compile(r":(\w+)")


def parse_type(field: FieldMetadata) -> List[str] | str:
    field_data: List[FieldType] = field.get("anyOf")
    if field_data:
        return [field_type.get("type") for field_type in field_data]

    return field.get("type", "string")


def parse_response_header_type(value: HTTPEncodable):
    if isinstance(value, int):
        return "integer"

    elif isinstance(value, float):
        return "number"

    elif isinstance(value, bool):
        return "boolean"

    else:
        return "string"


def parse_response_description(response: Any, status_code: int):
    response_description = f"Response for status code {status_code}"
    if response in Model.__subclasses__() and response.__doc__:
        response_description = response.__doc__

    return response_description


class EndpointParser:
    def __init__(
        self,
        path: str,
        methods: List[HTTPMethod],
        endpoint_metadata: ParsedEndpointMetadata,
        responses: Dict[
            int,
            type[HTML]
            | type[FileUpload]
            | type[Body]
            | type[Model]
            | type[dict]
            | type[list]
            | type[str]
            | type[bytes]
            | str
            | bytes,
        ],
        response_headers: Dict[str, Any] | None = None,
        headers: type[Headers] | None = None,
        parameters: type[Parameters] | None = None,
        query: type[Query] | None = None,
        body: (
            type[HTML]
            | type[FileUpload]
            | type[Body]
            | type[Model]
            | type[dict]
            | type[list]
            | type[str]
            | type[bytes]
            | str
            | bytes
            | None
        ) = None,
        required: List[
            Literal[
                "parameters",
                "headers",
                "query",
                "body",
            ]
        ]
        | None = None,
        default_tags: List[ParsedTag] | None = None,
    ) -> None:
        if default_tags is None:
            default_tags = []

        if required is None:
            required = []

        self.path = path
        self.path_params: Set[str] = set(path_pattern.findall(path))
        self.cleaned_path = "/".join(
            [
                ("{" + f"{segment.replace(':', '')}" + "}")
                if path_pattern.match(segment)
                else segment
                for segment in path.split("/")
            ]
        )

        self.methods = methods
        self.headers = headers
        self.parameters = parameters
        self.query = query
        self.body = body
        self.responses = responses
        self.metadata = endpoint_metadata
        self.required = required
        self.response_headers = response_headers
        self.default_tags = default_tags

        self.parsed_params: dict[str, Parameter] = {}
        self._content_type: str | None = None

        self.request_components: Dict[str, Any] = {}
        self.response_components: Dict[str, Any] = {}

    def parse(self) -> PathItem:
        param_types: List[ParamType] = [
            param_type
            for param_type in [self.parameters, self.query, self.headers]
            if param_type is not None
        ]

        for param in param_types:
            self.parsed_params.update(parse_param(param, self.path_params))

        operations: Dict[HTTPMethod, Operation] = {}

        for method in self.methods:
            operation_metadata: ParsedOperationMetadata = self.metadata.operations[
                method
            ]

            operation_tags: List[str] = []

            for tag in self.default_tags:
                if tag.value == operation_metadata.group_name:
                    operation_tags.append(tag.value)

            if operation_metadata.tags:
                operation_tags.extend(operation_metadata.tags)

            operations[method.lower()] = {
                "tags": operation_tags,
                "summary": operation_metadata.summary,
                "description": operation_metadata.description,
                "externalDocs": {
                    "description": operation_metadata.docs_description,
                    "url": operation_metadata.docs_url,
                }
                if operation_metadata.docs_url
                else None,
                "operationId": operation_metadata.name,
                "requestBody": self._parse_request_body(),
                "responses": self._parse_response(),
                "deprecated": operation_metadata.deprecated,
            }

        return {
            "description": self.metadata.description,
            "summary": self.metadata.summary,
            "parameters": list(self.parsed_params.values()),
            **operations,
        }

    def _parse_request_body(self) -> RequestBody:
        if self.body is None:
            return None

        content_type: str | None = None
        if self._content_type:
            content_type = self._content_type

        required = "body" in self.required

        if self.body in [FileUpload] or self.body in FileUpload.__subclasses__():
            if content_type is None:
                content_type = "application/octet-stream"

            body_config = self.body.defaults()

            body_content_type = body_config.get("content_type")
            body_encoding = body_config.get("encoding")

            if body_content_type:
                content_type = body_content_type

            encoding: str | None = None
            if body_encoding:
                encoding = body_encoding

            return {
                "description": self.body.__doc__,
                "content": {
                    content_type: {
                        "schema": {
                            "type": "string",
                            "format": "file",
                            "contentMediaType": content_type,
                            "contentEncoding": encoding,
                            "examples": self.body.model_json_schema(
                                ref_template=REF_TEMPLATE
                            ).get("examples"),
                        },
                    }
                },
                "required": required,
            }

        elif self.body in [HTML] or self.body in HTML.__subclasses__():
            if content_type is None:
                content_type = "text/html"

            return {
                "description": self.body.__doc__,
                "content": {
                    content_type: {
                        "schema": {
                            "type": "string",
                            "format": "html",
                            "contentMediaType": "text/html",
                            "contentEncoding": "utf-8",
                            "examples": self.body.model_json_schema(
                                ref_template=REF_TEMPLATE,
                            ).get("examples"),
                        },
                    }
                },
            }

        elif self.body == Body or self.body in Body.__subclasses__():
            content_type = "application/octet-stream"

        elif self.body in Model.__subclasses__() or self.body in [Model]:
            # schema: PropertySchema = msgspec.json.schema_components(
            #     [self.body],
            #     ref_template=REF_TEMPLATE.format(model=self.body.__name__),
            # )

            # schema_properties = schema.get(
            #     "$defs",
            #     {},
            # ).get(self.body.__name__, {})

            body_schema = self.body.model_json_schema(ref_template=REF_TEMPLATE)

            if content_type is None:
                content_type = "application/json"

            if defs := body_schema.get("$defs"):
                self.request_components.update(defs)

                del body_schema["$defs"]

            return {
                "description": self.body.__doc__,
                "content": {
                    content_type: {
                        "schema": {
                            **body_schema,
                            "contentMediaType": "application/json",
                            "contentEncoding": "utf-8",
                        },
                    }
                },
            }

        elif self.body == str or isinstance(self.body, str):
            content_type = "text/plain"

        elif self.body == bytes or isinstance(self.body, bytes):
            content_type = "application/octet-stream"

        elif self.body == dict or self.body == list:
            content_type = "application/json"

        elif self.body in dict.__subclasses__():
            content_type = "application/json"

        schema_type: SchemaType = "string"
        schema_format: str | None = None
        content_encoding: str | None = None

        match content_type:
            case "text/plain":
                schema_format = "string"

            case "application/octet-stream":
                schema_format = "binary"

            case "application/json":
                schema_format = "json"
                content_encoding = "utf-8"

        examples: List[str] = []
        if isinstance(self.body, str):
            examples.append(self.body)

        elif isinstance(self.body, bytes):
            examples.append(self.body.decode())

        return {
            "description": self.body.__doc__,
            "content": {
                content_type: {
                    "schema": {
                        "type": schema_type,
                        "format": schema_format,
                        "contentMediaType": content_type,
                        "contentEncoding": content_encoding,
                        "examples": examples,
                    },
                }
            },
        }

    def _parse_response(self) -> Dict[str, Response]:
        parsed_response_headers = {
            header_name: {
                "description": f"{header_name} header",
                "example": value,
                "required": True,
                "schema": {
                    "type": parse_response_header_type(value),
                },
            }
            for header_name, value in self.response_headers.items()
            if header_name.lower() not in ["content-type"]
        }

        responses: Dict[int, Dict[str, MediaType]] = {}
        for status_code, response in self.responses.items():
            responses[status_code] = self._parse_response_content(response)

        if self.responses.get(500) is None:
            self.responses[500] = InternalErrorSet

        if responses.get(500, {}).get("application/json") is None:
            responses[500] = self._parse_response_content(InternalErrorSet)

        if self.responses.get(422) is None:
            self.responses[422] = ValidationErrorGroup

        if self.response_components.get(400) is None:
            self.responses[400] = self._parse_response_content(BadRequestErrorSet)

        if responses.get(400, {}).get("application/json") is None:
            responses[400] = self._parse_response_content(BadRequestErrorSet)

        if responses.get(422, {}).get("application/json") is None:
            responses[422] = self._parse_response_content(ValidationErrorGroup)

        return {
            str(status_code): {
                "description": parse_response_description(response, status_code),
                "headers": parsed_response_headers,
                "content": responses[status_code],
            }
            for status_code, response in self.responses.items()
        }

    def _parse_response_content(
        self,
        response: (
            type[HTML]
            | type[FileUpload]
            | type[Body]
            | type[Model]
            | type[dict]
            | type[list]
            | type[str]
            | type[bytes]
            | str
            | bytes
        ),
    ) -> Dict[str, MediaType]:
        content_type: str | None = None
        if self._content_type:
            content_type = self._content_type

        if response in [FileUpload] or response in FileUpload.__subclasses__():
            if content_type is None:
                content_type = "application/octet-stream"

            response_config = response.defaults()
            response_content_type = response_config.get("content_type")
            response_encoding = response_config.get("encoding")

            if response_content_type:
                content_type = response_content_type

            encoding: str | None = None
            if response_encoding:
                encoding = response_encoding

            return {
                content_type: {
                    "schema": {
                        "type": "string",
                        "format": "file",
                        "contentMediaType": content_type,
                        "contentEncoding": encoding,
                    },
                },
            }

        elif response in [HTML] or response in HTML.__subclasses__():
            if content_type is None:
                content_type = "text/html"

            return {
                content_type: {
                    "schema": {
                        "type": "string",
                        "format": "html",
                        "contentMediaType": "text/html",
                        "contentEncoding": "utf-8",
                        "examples": response.model_json_schema(
                            ref_template=REF_TEMPLATE,
                        ).get("examples"),
                    },
                }
            }

        elif response in [Body] or response in Body.__subclasses__():
            content_type = "application/octet-stream"

        elif response in [Model] or response in Model.__subclasses__():
            response_schema = response.model_json_schema(ref_template=REF_TEMPLATE)

            if content_type is None:
                content_type = "application/json"

            if defs := response_schema.get("$defs"):
                self.response_components.update(defs)
                del response_schema["$defs"]

            return {
                "application/json": {
                    "schema": {
                        **response_schema,
                        "contentMediaType": "application/json",
                        "contentEncoding": "utf-8",
                    },
                }
            }

        elif response == str or isinstance(response, str):
            content_type = "text/plain"

        elif response == bytes or isinstance(response, bytes):
            content_type = "application/octet-stream"

        elif response in get_args(Json) or get_origin(response) in get_args(Json):
            content_type = "application/json"

        elif response in dict.__subclasses__():
            content_type = "application/json"

        schema_type: SchemaType = "string"
        schema_format: str | None = None
        content_encoding: str | None = None

        match content_type:
            case "text/plain":
                schema_format = "string"

            case "application/octet-stream":
                schema_format = "binary"

            case "application/json":
                schema_format = "json"
                content_encoding = "utf-8"

        examples: List[str] = []
        if isinstance(response, str):
            examples.append(response)

        elif isinstance(response, bytes):
            examples.append(response.decode())

        return {
            content_type: {
                "schema": {
                    "type": schema_type,
                    "format": schema_format,
                    "contentMediaType": content_type,
                    "contentEncoding": content_encoding,
                    "examples": examples,
                }
            }
        }



from __future__ import annotations

from typing import (
    Any,
    Dict,
    List,
    Literal,
    Optional,
)

from mkfst.middleware.base.response_context import ResponseContext
from mkfst.middleware.base.types import (
    CallHandler,
    Handler,
    MiddlewareHandler,
    MiddlewareResult,
)

from .bidirectional_wrapper import BidirectionalWrapper
from .call_wrapper import CallWrapper
from .types import MiddlewareType
from .unidirectional_wrapper import UnidirectionalWrapper


class Middleware:
    def __init__(
        self,
        name: str,
        middleware_type: MiddlewareType = MiddlewareType.UNIDIRECTIONAL_BEFORE,
        methods: Optional[
            List[
                Literal[
                    "GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"
                ]
            ]
        ] = None,
        response_headers: Dict[str, str] | None = None,
    ) -> None:
        if response_headers is None:
            response_headers = {}

        self.name = name
        self.methods = methods
        self.response_headers = response_headers
        self.middleware_type = middleware_type
        self.wraps = False

        self._wrapper_types = {
            MiddlewareType.BIDIRECTIONAL: BidirectionalWrapper,
            MiddlewareType.CALL: CallWrapper,
            MiddlewareType.UNIDIRECTIONAL_BEFORE: UnidirectionalWrapper,
            MiddlewareType.UNIDIRECTIONAL_AFTER: UnidirectionalWrapper,
        }

    def __call__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
    ) -> MiddlewareResult:
        raise NotImplementedError(
            "Err. __call__() should not be called on base Middleware class."
        )

    def wrap(self, handler: CallHandler):
        wrapper = self._wrapper_types.get(
            self.middleware_type,
            BidirectionalWrapper(
                self.name,
                handler,
                methods=self.methods,
                response_headers=self.response_headers,
                middleware_type=self.middleware_type,
            ),
        )(
            self.name,
            handler,
            methods=self.methods,
            response_headers=self.response_headers,
            middleware_type=self.middleware_type,
        )

        if isinstance(wrapper, BidirectionalWrapper):
            wrapper.pre = self.__pre__
            wrapper.post = self.__post__

        elif isinstance(wrapper, (CallWrapper, UnidirectionalWrapper)):
            wrapper.run = self.__run__

        wrapper.setup = self.__setup__
        self.wraps = wrapper.wraps

        return wrapper

    async def __setup__(self):
        pass

    async def __pre__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        raise NotImplementedError(
            "Err. - __pre__() is not implemented for base Middleware class."
        )

    async def __post__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        raise NotImplementedError(
            "Err. - __post__() is not implemented for base Middleware class."
        )

    async def __run__(
        self,
        context: ResponseContext | None = None,
        response: Any | None = None,
        handler: MiddlewareHandler | Handler | None = None,
    ) -> MiddlewareResult:
        raise NotImplementedError(
            "Err. - __run__() is not implemented for base Middleware class."
        )

    async def run(self, *args, **kwargs):
        raise NotImplementedError(
            "Err. - middleware() is not implemented for base Middleware class."
        )

    async def close(self):
        raise NotImplementedError(
            "Err. - close() is not implemented for base Middleware class."
        )

    def abort(self):
        raise NotImplementedError(
            "Err. - abort() is not implemented for base Middleware class."
        )



from .pages import get_redoc_html as get_redoc_html
from .pages import get_swagger_ui_html as get_swagger_ui_html
from .pages import (
    get_swagger_ui_oauth2_redirect_html as get_swagger_ui_oauth2_redirect_html,
)
from .parsed_api_metadata import ParsedAPIMetadata as ParsedAPIMetadata
from .parsed_api_server import ParsedAPIServer as ParsedAPIServer
from .parsed_api_server import ParsedServerVariable as ParsedServerVariable
from .parsed_docs import create_api_definition as create_api_definition
from .parsed_endpoint import ParsedEndpoint as ParsedEndpoint
from .parsed_endpoint_metadata import ParsedEndpointMetadata as ParsedEndpointMetadata
from .parsed_endpoint_metadata import ParsedOperationMetadata as ParsedOperationMetadata
from .parsed_tag import ParsedTag as ParsedTag



from typing import Any, Type
from mkfst.models import HTML, FileUpload, Body, Model


def get_content_type(return_type: Type[Any]):
    if return_type == FileUpload or return_type in FileUpload.__subclasses__():
        return "application/octet-stream"

    elif return_type == HTML or return_type in HTML.__subclasses__():
        return "text/html"

    elif return_type in [Body] or return_type in Body.__subclasses__():
        return "application/octet-stream"

    elif return_type in Model.__subclasses__():
        return "application/json"

    elif return_type == str or isinstance(return_type, str):
        return "text/plain"

    elif return_type == bytes or isinstance(return_type, bytes):
        return "application/octet-stream"

    else:
        return "application/json"



from typing import Any, Dict, List

from pydantic import BaseModel, RootModel, StrictStr


class PydanticValidationError(BaseModel):
    type: StrictStr | None = None
    loc: List[StrictStr] | None = None
    msg: StrictStr | None = None
    input: Dict[StrictStr, Any]
    url: StrictStr


class ValidationErrorGroup(RootModel):
    root: List[PydanticValidationError]



from __future__ import annotations

import asyncio
import datetime
import pathlib
import sys
import threading
from typing import Callable, Dict, TypeVar, Any

from mkfst.logging.models import Entry, Log, LogLevelName

from .logger_context import LoggerContext
from .retention_policy import RetentionPolicyConfig

T = TypeVar("T", bound=Entry)


class Logger:
    def __init__(self) -> None:
        self._contexts: Dict[str, LoggerContext] = {}
        self._watch_tasks: Dict[str, asyncio.Task] = {}

    def __getitem__(self, name: str):
        if self._contexts.get(name) is None:
            self._contexts[name] = LoggerContext(name=name)

        return self._contexts[name]

    def get_stream(
        self,
        name: str | None = None,
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ):
        if name is None:
            name = "default"

        filename: str | None = None
        directory: str | None = None

        if path:
            logfile_path = pathlib.Path(path)
            is_logfile = len(logfile_path.suffix) > 0

            filename = logfile_path.name if is_logfile else None
            directory = (
                str(logfile_path.parent.absolute())
                if is_logfile
                else str(logfile_path.absolute())
            )

        self._contexts[name] = LoggerContext(
            name=name,
            template=template,
            filename=filename,
            directory=directory,
            retention_policy=retention_policy,
            models=models,
            level=level,
        )

        return self._contexts[name].stream

    def configure(
        self,
        name: str | None = None,
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ):
        if name is None:
            name = "default"

        filename: str | None = None
        directory: str | None = None

        if path:
            logfile_path = pathlib.Path(path)
            is_logfile = len(logfile_path.suffix) > 0

            filename = logfile_path.name if is_logfile else None
            directory = (
                str(logfile_path.parent.absolute())
                if is_logfile
                else str(logfile_path.absolute())
            )

        self._contexts[name] = LoggerContext(
            name=name,
            template=template,
            filename=filename,
            directory=directory,
            retention_policy=retention_policy,
            models=models,
            level=level,
        )

    def context(
        self,
        name: str | None = None,
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        nested: bool = False,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ):
        if name is None:
            name = "default"

        filename: str | None = None
        directory: str | None = None

        if path:
            logfile_path = pathlib.Path(path)
            is_logfile = len(logfile_path.suffix) > 0

            filename = logfile_path.name if is_logfile else None
            directory = (
                str(logfile_path.parent.absolute())
                if is_logfile
                else str(logfile_path.absolute())
            )

        if self._contexts.get(name) is None:
            self._contexts[name] = LoggerContext(
                name=name,
                template=template,
                filename=filename,
                directory=directory,
                retention_policy=retention_policy,
                nested=nested,
                models=models,
                level=level,
            )

        else:
            self._contexts[name].name = name if name else self._contexts[name].name
            self._contexts[name].template = (
                template if template else self._contexts[name].template
            )
            self._contexts[name].filename = (
                filename if filename else self._contexts[name].filename
            )
            self._contexts[name].directory = (
                directory if directory else self._contexts[name].directory
            )
            self._contexts[name].retention_policy = (
                retention_policy
                if retention_policy
                else self._contexts[name].retention_policy
            )
            self._contexts[name].nested = nested

        return self._contexts[name]

    async def subscribe(
        self,
        logger: Logger,
        name: str | None = None,
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ):
        filename: str | None = None
        directory: str | None = None

        if name is None:
            name = "default"

        if path:
            logfile_path = pathlib.Path(path)
            is_logfile = len(logfile_path.suffix) > 0

            filename = logfile_path.name if is_logfile else None
            directory = (
                str(logfile_path.parent.absolute())
                if is_logfile
                else str(logfile_path.absolute())
            )

        if self._contexts.get(name) is None:
            self._contexts[name] = LoggerContext(
                name=name,
                template=template,
                filename=filename,
                directory=directory,
                retention_policy=retention_policy,
                models=models,
                level=level,
            )

            await self._contexts[name].stream.initialize()

        if logger._contexts.get(name) is None:
            logger._contexts[name] = LoggerContext(
                name=name,
            )

            await logger._contexts[name].stream.initialize()

        logger._contexts[name].stream._provider.subscribe(
            self._contexts[name].stream._consumer
        )

    async def log(
        self,
        entry: T,
        name: str | None = None,
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        filter: Callable[[T], bool] | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ):
        if name is None:
            name = "default"

        frame = sys._getframe(1)
        code = frame.f_code

        async with self.context(
            name=name,
            nested=True,
            models=models,
            level=level,
        ) as ctx:
            await ctx.log(
                Log(
                    level=entry.level,
                    entry=entry,
                    filename=code.co_filename,
                    function_name=code.co_name,
                    line_number=frame.f_lineno,
                    thread_id=threading.get_native_id(),
                    timestamp=datetime.datetime.now(datetime.UTC).isoformat(),
                ),
                template=template,
                path=path,
                retention_policy=retention_policy,
                filter=filter,
            )

    async def batch(
        self,
        *entries: T | Log[T],
        name: str | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ):
        if name is None:
            name = "default"

        frame = sys._getframe(1)
        code = frame.f_code

        async with self.context(
            name=name,
            nested=True,
            models=models,
            level=level,
        ) as ctx:
            await asyncio.gather(
                *[
                    ctx.put(
                        Log(
                            level=entry.level,
                            entry=entry,
                            filename=code.co_filename,
                            function_name=code.co_name,
                            line_number=frame.f_lineno,
                            thread_id=threading.get_native_id(),
                            timestamp=datetime.datetime.now(datetime.UTC).isoformat(),
                        ),
                    )
                    for entry in entries
                ]
            )

    async def put(
        self,
        entry: T | Log[T],
        name: str | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ):
        if name is None:
            name = "default"

        frame = sys._getframe(1)
        code = frame.f_code

        async with self.context(
            name=name,
            nested=True,
            models=models,
            level=level,
        ) as ctx:
            await ctx.put(
                Log(
                    level=entry.level,
                    entry=entry,
                    filename=code.co_filename,
                    function_name=code.co_name,
                    line_number=frame.f_lineno,
                    thread_id=threading.get_native_id(),
                    timestamp=datetime.datetime.now(datetime.UTC).isoformat(),
                ),
            )

    def watch(
        self,
        name: str | None = None,
        filter: Callable[[T], bool] | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
    ):
        if name is None:
            name = "default"

        if self._watch_tasks.get(name):
            try:
                self._watch_tasks[name].cancel()

            except (asyncio.CancelledError, asyncio.InvalidStateError):
                pass

        self._watch_tasks[name] = asyncio.create_task(
            self._watch(
                name,
                filter=filter,
                models=models,
            )
        )

    async def _watch(
        self,
        name: str,
        filter: Callable[[T], bool] | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
    ):
        async with self.context(
            name=name,
            nested=True,
            models=models,
        ) as ctx:
            async for log in ctx.get(filter=filter):
                await ctx.log(log)

    async def stop_watch(self, name: str | None = None):
        if name is None:
            name = "default"

        if (context := self._contexts.get(name)) and (
            watch_task := self._watch_tasks.get(name)
        ):
            await context.stream.close(shutdown_subscribed=True)

            try:
                await watch_task

            except (
                asyncio.CancelledError,
                asyncio.InvalidStateError,
            ):
                pass

    async def close(self):
        if len(self._watch_tasks) > 0:
            await asyncio.gather(*[self.stop_watch(name) for name in self._watch_tasks])

        shutdown_subscribed = (
            len(
                [
                    context
                    for context in self._contexts.values()
                    if context.stream.has_active_subscriptions
                ]
            )
            > 0
        )

        contexts_count = len(self._contexts)

        if contexts_count > 0:
            await asyncio.gather(
                *[
                    context.stream.close(shutdown_subscribed=shutdown_subscribed)
                    for context in self._contexts.values()
                ]
            )

    def abort(self):
        for context in self._contexts.values():
            context.stream.abort()



import asyncio
import datetime
import io
import functools
import os
import pathlib
import sys
import threading
import uuid
from collections import defaultdict
from typing import (
    Callable,
    Dict,
    List,
    TypeVar,
    Any,
)

import msgspec
import zstandard

from mkfst.logging.config.logging_config import LoggingConfig
from mkfst.logging.models import Entry, Log, LogLevel, LogLevelName
from mkfst.logging.queue import (
    ConsumerStatus,
    LogConsumer,
    LogProvider,
)
from mkfst.logging.snowflake import SnowflakeGenerator

from .protocol import LoggerProtocol
from .retention_policy import (
    RetentionPolicy,
    RetentionPolicyConfig,
)
from mkfst.logging.config.stream_type import StreamType

T = TypeVar("T", bound=Entry)

try:
    import uvloop as uvloop

    has_uvloop = True

except Exception:
    has_uvloop = False


def patch_transport_close(
    transport: asyncio.Transport,
    loop: asyncio.AbstractEventLoop,
):
    def close(*args, **kwargs):
        try:
            transport.close()

        except Exception:
            pass

    return close


class LoggerStream:
    def __init__(
        self,
        name: str | None = None,
        template: str | None = None,
        filename: str | None = None,
        directory: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        models: dict[
            str,
            tuple[
                type[T],
                dict[str, Any],
            ],
        ]
        | None = None,
        level: LogLevelName | None = None,
    ) -> None:
        if name is None:
            name = "default"

        self._name = name
        self._default_template = template
        self._default_logfile = filename
        self._default_log_directory = directory

        self._default_retention_policy = retention_policy
        if retention_policy:
            self._default_retention_policy = RetentionPolicy(retention_policy)
            self._default_retention_policy.parse()

        self._init_lock = asyncio.Lock()
        self._stream_writers: Dict[StreamType, asyncio.StreamWriter] = {}
        self._loop: asyncio.AbstractEventLoop | None = None
        self._generator: SnowflakeGenerator | None = None
        self._compressor: zstandard.ZstdCompressor | None = None

        self._files: Dict[str, io.FileIO] = {}
        self._file_locks: Dict[str, asyncio.Lock] = defaultdict(asyncio.Lock)
        self._cwd: str | None = None
        self._default_logfile_path: str | None = None

        self._retention_policies: Dict[str, RetentionPolicy] = {}

        self._config = LoggingConfig(
            level=level,
        )
        self._initialized: bool = False
        self._consumer: LogConsumer | None = None
        self._provider: LogProvider | None = None
        self._initialized: bool = False
        self._closed = False
        self._stderr: io.TextIOBase | None = None
        self._stdout: io.TextIOBase | None = None
        self._transports: List[asyncio.Transport] = []

        self._models: Dict[str, Callable[..., Entry]] = {}
        self._queue: asyncio.Queue[asyncio.Future] = asyncio.Queue()

        if models is None:
            models = {}

        for name, config in models.items():
            model, defaults = config

            self._models[name] = (model, defaults)

        self._models.update({"default": (Entry, {"level": LogLevel.INFO})})

        self._log_levels_map = {
            LogLevel.TRACE: 0,
            LogLevel.DEBUG: 1,
            LogLevel.INFO: 2,
            LogLevel.WARN: 3,
            LogLevel.ERROR: 4,
            LogLevel.CRITICAL: 5,
            LogLevel.FATAL: 6,
        }

        self._level_value = self._log_levels_map[self._config.level]

    @property
    def has_active_subscriptions(self):
        if self._provider is None:
            return False

        return self._provider.subscriptions_count > 0

    async def initialize(self) -> asyncio.StreamWriter:
        async with self._init_lock:
            if self._initialized:
                return

            if self._generator is None:
                self._generator = SnowflakeGenerator(
                    (uuid.uuid1().int + threading.get_native_id()) >> 64
                )

            if self._compressor is None:
                self._compressor = zstandard.ZstdCompressor()

            if self._loop is None:
                self._loop = asyncio.get_event_loop()

            if self._consumer is None:
                self._consumer = LogConsumer()

            if self._provider is None:
                self._provider = LogProvider()

            if self._stdout is None or self._stdout.closed:
                self._stdout = await self._dup_stdout()

            if self._stderr is None or self._stderr.closed:
                self._stderr = await self._dup_stderr()

            if self._stream_writers.get(StreamType.STDOUT) is None:
                transport, protocol = await self._loop.connect_write_pipe(
                    lambda: LoggerProtocol(), self._stdout
                )

                try:
                    if has_uvloop:
                        transport.close = patch_transport_close(transport, self._loop)

                except Exception:
                    pass

                self._stream_writers[StreamType.STDOUT] = asyncio.StreamWriter(
                    transport,
                    protocol,
                    None,
                    self._loop,
                )

            if self._stream_writers.get(StreamType.STDERR) is None:
                transport, protocol = await self._loop.connect_write_pipe(
                    lambda: LoggerProtocol(), self._stderr
                )

                try:
                    if has_uvloop:
                        transport.close = patch_transport_close(transport, self._loop)

                except Exception:
                    pass

                self._stream_writers[StreamType.STDERR] = asyncio.StreamWriter(
                    transport,
                    protocol,
                    None,
                    self._loop,
                )

            self._initialized = True

    async def open_file(
        self,
        filename: str,
        directory: str | None = None,
        is_default: bool = False,
        retention_policy: RetentionPolicyConfig | None = None,
    ):
        if self._cwd is None:
            self._cwd = await self._loop.run_in_executor(
                None,
                os.getcwd,
            )

        logfile_path = self._to_logfile_path(filename, directory=directory)
        await self._file_locks[logfile_path].acquire()

        await self._loop.run_in_executor(
            None,
            self._open_file,
            logfile_path,
        )

        file_lock = self._file_locks[logfile_path]

        if file_lock.locked():
            file_lock.release()

        if retention_policy and self._retention_policies.get(logfile_path) is None:
            policy = RetentionPolicy(retention_policy)
            policy.parse()

            self._retention_policies[logfile_path] = policy

        if is_default:
            self._default_logfile_path = logfile_path

    def _open_file(
        self,
        logfile_path: str,
    ):
        resolved_path = pathlib.Path(logfile_path).absolute().resolve()
        logfile_directory = str(resolved_path.parent)
        path = str(resolved_path)

        if not os.path.exists(logfile_directory):
            os.makedirs(logfile_directory)

        if not os.path.exists(path):
            resolved_path.touch()

        self._files[logfile_path] = open(path, "ab+")

    async def _rotate(self, logfile_path: str, retention_policy: RetentionPolicy):
        await self._file_locks[logfile_path].acquire()
        await self._loop.run_in_executor(
            None,
            self._rotate_logfile,
            retention_policy,
            logfile_path,
        )

        file_lock = self._file_locks[logfile_path]

        if file_lock.locked():
            file_lock.release()

    def _get_logfile_metadata(self, logfile_path: str) -> Dict[str, float]:
        resolved_path = pathlib.Path(logfile_path)

        logfile_metadata_path = os.path.join(
            str(resolved_path.parent.absolute().resolve()), ".logging.json"
        )

        if os.path.exists(logfile_metadata_path):
            metadata_file = open(logfile_metadata_path, "+rb")
            return msgspec.json.decode(metadata_file.read())

        return {}

    def _update_logfile_metadata(
        self,
        logfile_path: str,
        logfile_metadata: Dict[str, float],
    ):
        resolved_path = pathlib.Path(logfile_path)

        logfile_metadata_path = os.path.join(
            str(resolved_path.parent.absolute().resolve()), ".logging.json"
        )

        with open(logfile_metadata_path, "+wb") as metadata_file:
            metadata_file.write(msgspec.json.encode(logfile_metadata))

    def _rotate_logfile(
        self,
        retention_policy: RetentionPolicy,
        logfile_path: str,
    ):
        resolved_path = pathlib.Path(logfile_path)
        path = str(resolved_path.absolute().resolve())

        logfile_metadata = self._get_logfile_metadata(logfile_path)

        current_time = datetime.datetime.now(datetime.UTC)
        current_timestamp = current_time.timestamp()

        created_time = logfile_metadata.get(
            logfile_path,
            current_timestamp,
        )

        archived_filename = f"{resolved_path.stem}_{current_timestamp}_archived.zst"
        logfile_data = b""

        if (
            retention_policy.matches_policy(
                {
                    "file_age": (
                        current_time
                        - datetime.datetime.fromtimestamp(created_time, datetime.UTC)
                    ).seconds,
                    "file_size": os.path.getsize(logfile_path),
                    "logfile_path": resolved_path,
                }
            )
            is False
        ):
            self._files[logfile_path].close()

            with open(logfile_path, "rb") as logfile:
                logfile_data = logfile.read()

        if len(logfile_data) > 0:
            archive_path = os.path.join(
                str(resolved_path.parent.absolute().resolve()),
                archived_filename,
            )

            with open(archive_path, "wb") as archived_file:
                archived_file.write(self._compressor.compress(logfile_data))

            self._files[logfile_path] = open(path, "wb+")
            created_time = current_timestamp

        logfile_metadata[logfile_path] = created_time

        self._update_logfile_metadata(logfile_path, logfile_metadata)

    async def close(self, shutdown_subscribed: bool = False):
        if self._consumer:
            self._consumer.stop()

        if shutdown_subscribed:
            await self._provider.signal_shutdown()

        if self._consumer and (
            self._consumer.status
            in [
                ConsumerStatus.RUNNING,
                ConsumerStatus.CLOSING,
            ]
            and self._consumer.pending
        ):
            await self._consumer.wait_for_pending()

        while not self._queue.empty():
            task = self._queue.get_nowait()
            await task

        await asyncio.gather(
            *[self._close_file(logfile_path) for logfile_path in self._files]
        )

        await asyncio.gather(
            *[writer.drain() for writer in self._stream_writers.values()]
        )

        self._initialized = False

    def abort(self):
        for logfile_path in self._files:
            if (logfile := self._files.get(logfile_path)) and logfile.closed is False:
                try:
                    logfile.close()

                except Exception:
                    pass

        if self._consumer:
            self._consumer.abort()

        while not self._queue.empty():
            task = self._queue.get_nowait()
            task.set_result(None)

    async def close_file(
        self,
        filename: str,
        directory: str | None = None,
    ):
        if self._cwd is None:
            self._cwd = await self._loop.run_in_executor(None, os.getcwd)

        logfile_path = self._to_logfile_path(filename, directory=directory)
        await self._close_file(logfile_path)

    async def _close_file(self, logfile_path: str):
        if file_lock := self._file_locks.get(logfile_path):
            if file_lock.locked():
                file_lock.release()

            await file_lock.acquire()
            await self._loop.run_in_executor(
                None,
                self._close_file_at_path,
                logfile_path,
            )

            if file_lock.locked():
                file_lock.release()

    def _close_file_at_path(self, logfile_path: str):
        if (logfile := self._files.get(logfile_path)) and logfile.closed is False:
            logfile.close()

    def _to_logfile_path(
        self,
        filename: str,
        directory: str | None = None,
    ):
        filename_path = pathlib.Path(filename)

        assert filename_path.suffix == ".json", (
            "Err. - file must be JSON file for logs."
        )

        if self._config.directory:
            directory = self._config.directory

        elif directory is None:
            directory: str = os.path.join(self._cwd)

        logfile_path: str = os.path.join(directory, filename_path)

        return logfile_path

    async def _dup_stdout(self):
        stdout_fileno = await self._loop.run_in_executor(None, sys.stderr.fileno)

        stdout_dup = await self._loop.run_in_executor(
            None,
            os.dup,
            stdout_fileno,
        )

        return await self._loop.run_in_executor(
            None, functools.partial(os.fdopen, stdout_dup, mode=sys.stdout.mode)
        )

    async def _dup_stderr(self):
        stderr_fileno = await self._loop.run_in_executor(None, sys.stderr.fileno)

        stderr_dup = await self._loop.run_in_executor(
            None,
            os.dup,
            stderr_fileno,
        )

        return await self._loop.run_in_executor(
            None, functools.partial(os.fdopen, stderr_dup, mode=sys.stderr.mode)
        )

    def schedule(
        self,
        entry: T,
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        filter: Callable[[T], bool] | None = None,
    ):
        self._queue.put_nowait(
            asyncio.ensure_future(
                self.log(
                    entry,
                    template=template,
                    path=path,
                    retention_policy=retention_policy,
                    filter=filter,
                )
            )
        )

    async def log_prepared_batch(
        self,
        model_messages: dict[str, list[str]],
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        filter: Callable[[T], bool] | None = None,
    ):
        entries = [
            self._to_entry(
                message,
                name,
            )
            for name, messages in model_messages.items()
            for message in messages
        ]

        if len(entries) > 0:
            await asyncio.gather(
                *[
                    self.log(
                        entry,
                        template=template,
                        path=path,
                        retention_policy=retention_policy,
                        filter=filter,
                    )
                    for entry in entries
                ],
                return_exceptions=True,
            )

    async def batch(
        self,
        entries: list[T],
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        filter: Callable[[T], bool] | None = None,
    ):
        if len(entries) > 0:
            await asyncio.gather(
                *[
                    self.log(
                        entry,
                        template=template,
                        path=path,
                        retention_policy=retention_policy,
                        filter=filter,
                    )
                    for entry in entries
                ],
                return_exceptions=True,
            )

    async def log_prepared(
        self,
        message: str,
        name: str = "default",
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        filter: Callable[[T], bool] | None = None,
    ):
        entry = self._to_entry(message, name)
        if self._log_levels_map[entry.level] < self._level_value:
            return

        await self.log(
            entry,
            template=template,
            path=path,
            retention_policy=retention_policy,
            filter=filter,
        )

    async def log(
        self,
        entry: T,
        template: str | None = None,
        path: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        filter: Callable[[T], bool] | None = None,
    ):
        filename: str | None = None
        directory: str | None = None

        if self._log_levels_map[entry.level] < self._level_value:
            return

        if path:
            logfile_path = pathlib.Path(path)
            is_logfile = len(logfile_path.suffix) > 0

            filename = logfile_path.name if is_logfile else None
            directory = (
                str(logfile_path.parent.absolute())
                if is_logfile
                else str(logfile_path.absolute())
            )

        if template is None:
            template = self._default_template

        if filename is None:
            filename = self._default_logfile

        if directory is None:
            directory = self._default_log_directory

        if retention_policy is None:
            retention_policy = self._default_retention_policy

        if filename or directory:
            await self._log_to_file(
                entry,
                filename=filename,
                directory=directory,
                retention_policy=retention_policy,
                filter=filter,
            )

        else:
            await self._log(
                entry,
                template=template,
                filter=filter,
            )

    def _to_entry(
        self,
        message: str,
        name: str,
    ) -> Entry:
        model, defaults = self._models.get(name, self._models.get("default"))

        return model(message=message, **defaults)

    async def _log(
        self,
        entry_or_log: T | Log[T],
        template: str | None = None,
        filter: Callable[[T], bool] | None = None,
    ):
        entry: Entry = None
        if isinstance(entry_or_log, Log):
            entry = entry_or_log.entry

        else:
            entry = entry_or_log

        if self._config.enabled(self._name, entry.level) is False:
            return

        if filter and filter(entry) is False:
            return

        if self._initialized is None:
            await self.initialize()

        stream_writer = self._stream_writers[self._config.output]

        if stream_writer.is_closing():
            return

        if template is None:
            template = "{timestamp} - {level} - {thread_id} - {filename}:{function_name}.{line_number} - {message}"

        if isinstance(entry_or_log, Log):
            log_file = entry_or_log.filename
            line_number = entry_or_log.line_number
            function_name = entry_or_log.function_name

        else:
            log_file, line_number, function_name = self._find_caller()

        if self._stdout is None or self._stdout.closed:
            self._stdout = await self._dup_stdout()

        if self._stderr is None or self._stderr.closed:
            self._stderr = await self._dup_stderr()

        try:
            stream_writer.write(
                entry.to_template(
                    template,
                    context={
                        "filename": log_file,
                        "function_name": function_name,
                        "line_number": line_number,
                        "thread_id": threading.get_native_id(),
                        "timestamp": datetime.datetime.now(datetime.UTC).isoformat(),
                    },
                ).encode()
                + b"\n"
            )

            await stream_writer.drain()

        except Exception as err:
            error_template = "{timestamp} - {level} - {thread_id}.{filename}:{function_name}.{line_number} - {error}"

            if self._stderr.closed is False:
                await self._loop.run_in_executor(
                    None,
                    self._stderr.write,
                    entry.to_template(
                        error_template,
                        context={
                            "filename": log_file,
                            "function_name": function_name,
                            "line_number": line_number,
                            "error": str(err),
                            "thread_id": threading.get_native_id(),
                            "timestamp": datetime.datetime.now(
                                datetime.UTC
                            ).isoformat(),
                        },
                    ),
                )

    async def _log_to_file(
        self,
        entry_or_log: T | Log[T],
        filename: str | None = None,
        directory: str | None = None,
        retention_policy: RetentionPolicyConfig | None = None,
        filter: Callable[[T], bool] | None = None,
    ):
        entry: Entry = None
        if isinstance(entry_or_log, Log):
            entry = entry_or_log.entry

        else:
            entry = entry_or_log

        if self._config.enabled(self._name, entry.level) is False:
            return

        if filter and filter(entry) is False:
            return

        if self._cwd is None:
            self._cwd = await self._loop.run_in_executor(
                None,
                os.getcwd,
            )

        if filename and directory:
            logfile_path = self._to_logfile_path(
                filename,
                directory=directory,
            )

        elif self._default_logfile_path:
            logfile_path = self._default_logfile_path

        else:
            filename = "logs.json"
            directory = os.path.join(self._cwd, "logs")
            logfile_path = os.path.join(directory, filename)

        if self._files.get(logfile_path) is None or self._files[logfile_path].closed:
            await self.open_file(
                filename,
                directory=directory,
            )

        if retention_policy:
            self._retention_policies[logfile_path] = retention_policy

        if retention_policy := self._retention_policies.get(logfile_path):
            await self._rotate(
                logfile_path,
                retention_policy,
            )

        if isinstance(entry_or_log, Log):
            log_file = entry_or_log.filename
            line_number = entry_or_log.line_number
            function_name = entry_or_log.function_name

            log = entry_or_log

        else:
            log_file, line_number, function_name = self._find_caller()

            log = Log(
                level=entry.level,
                entry=entry,
                filename=log_file,
                function_name=function_name,
                line_number=line_number,
            )

        try:
            file_lock = self._file_locks[logfile_path]
            await file_lock.acquire()

            await self._loop.run_in_executor(
                None,
                self._write_to_file,
                log,
                logfile_path,
            )

            if file_lock.locked():
                file_lock.release()

            await asyncio.sleep(0)

        except Exception as err:
            file_lock = self._file_locks[logfile_path]

            if file_lock.locked():
                file_lock.release()

            error_template = "{timestamp} - {level} - {thread_id}.{filename}:{function_name}.{line_number} - {error}"

            if self._stderr.closed is False:
                await self._loop.run_in_executor(
                    None,
                    self._stderr.write,
                    entry.to_template(
                        error_template,
                        context={
                            "filename": log_file,
                            "function_name": function_name,
                            "line_number": line_number,
                            "error": str(err),
                            "thread_id": threading.get_native_id(),
                            "timestamp": datetime.datetime.now(
                                datetime.UTC
                            ).isoformat(),
                        },
                    ),
                )

    def _write_to_file(
        self,
        log: Log,
        logfile_path: str,
    ):
        try:
            if (logfile := self._files.get(logfile_path)) and (logfile.closed is False):
                logfile.write(msgspec.json.encode(log) + b"\n")
                logfile.flush()

        except Exception:
            pass

    def _find_caller(self):
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        frame = sys._getframe(3)
        code = frame.f_code

        return (
            code.co_filename,
            frame.f_lineno,
            code.co_name,
        )

    async def get(self, filter: Callable[[T], bool] | None = None):
        async for log in self._consumer.iter_logs(filter=filter):
            yield log

    async def put(
        self,
        entry: T | Log[T],
    ):
        if not isinstance(entry, Log):
            frame = sys._getframe(1)
            code = frame.f_code
            entry = Log(
                level=entry.level,
                entry=entry,
                filename=code.co_filename,
                function_name=code.co_name,
                line_number=frame.f_lineno,
                thread_id=threading.get_native_id(),
                timestamp=datetime.datetime.now(datetime.UTC).isoformat(),
            )

        await self._provider.put(entry)



import asyncio
import math
import os
import statistics
from typing import List, Union

import psutil

from mkfst.models.http import Limit

from .base_limiter import BaseLimiter


class CPUAdaptiveLimiter(BaseLimiter):
    __slots__ = (
        "max_rate",
        "time_period",
        "_rate_per_sec",
        "_level",
        "_waiters",
        "_loop",
        "_last_check",
        "_cpu_limit",
        "_current_time",
        "_previous_count",
        "_current_cpu",
        "_max_queue",
        "_sample_task",
        "_running",
        "_process",
        "_max_fast_backoff",
        "_min_backoff",
        "_history",
    )

    def __init__(self, limit: Limit) -> None:
        super().__init__(
            limit.max_requests, limit.period, reject_requests=limit.reject_requests
        )

        cpu_limit = limit.cpu_limit
        if cpu_limit is None:
            cpu_limit = 50

        self._cpu_limit = cpu_limit
        self._backoff = limit.backoff
        self._min_backoff = self._backoff
        self._max_fast_backoff = math.ceil(self._backoff * 10)
        self._max_backoff = math.ceil(self._max_fast_backoff * 10)
        self._last_check = self._loop.time()
        self._current_time = self._loop.time()
        self._previous_count = limit.max_requests

        self._history: List[float] = []

        self._max_queue = limit.max_requests
        self._sample_task: Union[asyncio.Task, None] = None
        self._running = False
        self._activate_limit = False
        self._process = psutil.Process(os.getpid())

        self._current_cpu = self._process.cpu_percent()
        self._history.append(self._current_cpu)

    def has_capacity(self, amount: float = 1) -> bool:
        elapsed = self._loop.time() - self._last_check

        self._backoff = max(
            self._backoff - (1 / self._rate_per_sec * elapsed), self._min_backoff
        )

        if (self._loop.time() - self._current_time) > self.time_period:
            self._current_time = (
                math.floor(self._loop.time() / self.time_period) * self.time_period
            )

            self._previous_count = self._level
            self._level = 0

        self._rate_per_sec = (
            self._previous_count
            * (self.time_period - (self._loop.time() - self._current_time))
            / self.time_period
        ) + (self._level + amount)

        if self._rate_per_sec < self.max_rate:
            for fut in self._waiters.values():
                if not fut.done():
                    fut.set_result(True)
                    break

        self._last_check = self._loop.time()

        return self._rate_per_sec <= self.max_rate

    async def acquire(
        self,
        amount: float = 1,
    ):      
        if not self._running:
            self._running = True
            self._sample_task = asyncio.create_task(self._sample_cpu())

        if amount > self.max_rate:
            raise ValueError("Can't acquire more than the maximum capacity")

        task = asyncio.current_task(loop=self._loop)

        assert task is not None

        rejected = False

        while not self.has_capacity(amount) or self._activate_limit:
            fut = self._loop.create_future()
            try:
                self._waiters[task] = fut

                await asyncio.wait_for(asyncio.shield(fut), timeout=self._backoff)

                if self._activate_limit:
                    await asyncio.sleep(self._backoff)
                    self._max_fast_backoff = min(
                        self._max_fast_backoff + (1 / math.sqrt(self._rate_per_sec)),
                        self._max_backoff,
                    )

            except asyncio.TimeoutError:
                pass

            fut.cancel()

            rejected = True

        self._backoff = min(self._backoff * 2, self._max_fast_backoff)
        self._waiters.pop(task, None)
        self._level += amount

        return rejected

    async def _sample_cpu(self):
        while self._running:
            self._current_cpu = self._process.cpu_percent()
            self._history.append(self._current_cpu)

            elapsed = self._loop.time() - self._last_check

            if elapsed > self.time_period:
                self._history.pop(0)

            if self._current_cpu >= self._cpu_limit:
                self._activate_limit = True

            elif statistics.median(self._history) < self._cpu_limit:
                self._activate_limit = False
                self._max_fast_backoff = max(
                    self._max_fast_backoff - (1 / self._rate_per_sec), self._min_backoff
                )

            await asyncio.sleep(0.1)

    async def close(self):
        self._running = False

        self._sample_task.cancel()
        if not self._sample_task.cancelled():
            try:
                await self._sample_task

            except (asyncio.CancelledError, asyncio.InvalidStateError):
                pass

    def abort(self):
        self._running = False

        try:
            self._sample_task.cancel()
        except (asyncio.CancelledError, asyncio.InvalidStateError):
            pass



from enum import Enum
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Tuple,
    Union,
)

from .response_context import ResponseContext

PositionalArgs = Tuple[Any, ...]
KeywordArgs = Dict[str, Any]

class MiddlewareType(Enum):
    BIDIRECTIONAL = "BIDIRECTIONAL"
    CALL = "CALL"
    UNIDIRECTIONAL_BEFORE = "UNIDIRECTIONAL_BEFORE"
    UNIDIRECTIONAL_AFTER = "UNIDIRECTIONAL_AFTER"

MiddlewareResult = Tuple[ResponseContext, Any]
RunResult = Tuple[MiddlewareResult, bool]

RequestHandler = Callable[
    ..., 
    Coroutine[Any, Any, Any]
]

WrappedHandler = Callable[
    ..., 
    Coroutine[Any, Any, Any]
]

CallHandler = Callable[
    ..., 
    Coroutine[Any, Any, Any]
]

MiddlewareHandler = Callable[
    ..., 
    Coroutine[Any, Any, RunResult]
]


BidirectionalMiddlewareHandler = Callable[
    [
        ResponseContext,
        Any,
        CallHandler | MiddlewareHandler
    ],
    Coroutine[
        Any, 
        Any, 
        Tuple[
            MiddlewareResult, 
            bool
        ]
    ],
]


Handler = Union[RequestHandler, WrappedHandler]



from __future__ import annotations
from typing import Callable, List, Literal, Optional
import msgspec
from ipaddress import IPv4Address, IPv6Address
from mkfst.env.memory_parser import MemoryParser
from mkfst.env.time_parser import TimeParser

HTTPMethod = Literal[
    "GET", "HEAD", "OPTIONS", "POST", "PUT", "PATCH", "DELETE", "TRACE"
]


class Limit(msgspec.Struct):
    max_requests: int = 1000
    min_requests: int = 100
    request_period: str = "1s"
    reject_requests: bool = True
    request_backoff: str = "1s"
    cpu_limit: float | int | None = None
    memory_limit: str = None
    limiter_type: Literal[
        "adaptive",
        "cpu-adaptive",
        "leaky-bucket",
        "rate-adaptive",
        "sliding-window",
        "token-bucket",
    ] = "sliding-window"
    limit_key: Optional[
        Callable[
            [
                str,
                str,
                IPv4Address | IPv6Address,
            ],
            str,
        ]
    ] = None
    rules: Optional[
        List[
            Callable[
                [
                    str,
                    str,
                    IPv4Address | IPv6Address,
                ],
                bool,
            ]
        ]
    ] = None

    @property
    def backoff(self):
        return TimeParser(self.request_backoff).time

    @property
    def period(self):
        return TimeParser(self.request_period).time

    @property
    def memory(self):
        return MemoryParser(self.memory_limit).megabytes(accuracy=4)

    def get_key(
        self,
        path: str,
        method: str,
        ip_address: IPv4Address | IPv6Address,
        default: str = "default",
    ):
        if self.limit_key is None:
            return default

        return self.limit_key(path, method, ip_address)

    def matches(self, path, method, ip_address: IPv4Address | IPv6Address):
        if self.rules is None:
            return True

        matches_rules = False

        for rule in self.rules:
            matches_rules = rule(
                path,
                method,
                ip_address,
            )

        return matches_rules



from __future__ import annotations

import asyncio
import ipaddress
import re
import socket
from collections import defaultdict, deque
from functools import lru_cache
from typing import (
    Any,
    Callable,
    Deque,
    Dict,
    List,
    Literal,
    Optional,
    Tuple,
    Union,
    Iterable,
)

import orjson
import psutil

from mkfst.connection.base.connection_type import ConnectionType
from mkfst.env import Env
from mkfst.logging import Logger, LogLevel
from mkfst.middleware.base.response_context import ResponseContext
from mkfst.models.http import (
    HTTPResponse,
    parse_response,
    Model,
)
from mkfst.models.logging import Event, Response, Request
from mkfst.rate_limiting import Limiter

from .fabricator import Fabricator
from .mercury_sync_tcp_connection import MercurySyncTCPConnection
from .protocols.receive_buffer import ReceiveBuffer
from .protocols.patterns import (
    request_line,
    header_field,
    request_target,
    method as method_re,
)
from .router import Router
from .protocols.validate import validate

Handler = Callable[..., Tuple[Any, int]]


request_line_re = re.compile(request_line.encode("ascii"))
header_field_re = re.compile(header_field.encode("ascii"))
method_re = re.compile(method_re.encode("ascii"))
request_target_re = re.compile(request_target.encode("ascii"))

obs_fold_re = re.compile(rb"[ \t]+")


def _obsolete_line_fold(lines: Iterable[bytes]) -> Iterable[bytes]:
    it = iter(lines)
    last: Optional[bytes] = None
    for line in it:
        match = obs_fold_re.match(line)
        if match:
            if last is None:
                raise Exception("continuation line at start of headers")
            if not isinstance(last, bytearray):
                # Cast to a mutable type, avoiding copy on append to ensure O(n) time
                last = bytearray(last)
            last += b" "
            last += line[match.end() :]
        else:
            if last is not None:
                yield last
            last = line
    if last is not None:
        yield last


class MercurySyncHTTPConnection(MercurySyncTCPConnection):
    def __init__(
        self,
        host: str,
        port: int,
        instance_id: int,
        env: Env,
        upgrade_port: int | None = None,
    ) -> None:
        super().__init__(host, port, instance_id, env)

        self.env = env
        self._waiters: dict[str, asyncio.Future] = defaultdict(asyncio.Future)
        self._connections: Dict[str, List[asyncio.Transport]] = defaultdict(list)
        self._http_socket: Union[socket.socket, None] = None
        self._hostnames: Dict[Tuple[str, int], str] = {}
        self._max_concurrency = env.MERCURY_SYNC_MAX_CONCURRENCY

        self.connection_type = ConnectionType.HTTP
        self._is_server = env.MERCURY_SYNC_USE_HTTP_SERVER
        self._use_encryption = env.MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION

        self._supported_handlers: Dict[str, Dict[str, str]] = defaultdict(dict)
        self._request_parsers: Dict[Model, Tuple[Callable[[Any], str], int]] = {}
        self._response_parsers: Dict[Model, Tuple[Callable[[Any], str], int]] = {}
        self._response_headers: Dict[str, Dict[str, Any]] = {}

        self._middleware_enabled: Dict[str, bool] = {}

        self._limiter = Limiter(env)

        self._backoff_sem: Union[asyncio.Semaphore, None] = None

        rate_limit_strategy = env.MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY
        self._rate_limiting_enabled = rate_limit_strategy != "none"
        self._rate_limiting_backoff_rate = env.MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE

        self._initial_cpu = psutil.cpu_percent()
        self._path_param_pattern = re.compile(r"(:\w+)")
        self.routes = Router()
        self.match_routes = {}

        self.fabricators: Dict[str, Fabricator] = {}
        self._logger = Logger()
        self._cache: dict[bytes, tuple[bytes, int, float]] = {}
        self._max_request_cache_size = env.MERCURY_SYNC_MAX_REQUEST_CACHE_SIZE
        self._cache_purge_lock: asyncio.Lock | None = None
        self._request_caching_enabled = env.MERCURY_SYNC_ENABLE_REQUEST_CACHING
        self._verify_cert = env.MERCURY_SYNC_VERIFY_SSL_CERT
        self._upgrade_port = upgrade_port
        self.waiting_for_data: asyncio.Event = asyncio.Event()

    def from_env(self, env: Env):
        super().from_env(env)

        self._max_concurrency = env.MERCURY_SYNC_MAX_CONCURRENCY
        self._is_server = env.MERCURY_SYNC_USE_HTTP_SERVER
        self._use_encryption = env.MERCURY_SYNC_USE_HTTP_MSYNC_ENCRYPTION
        rate_limit_strategy = env.MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY
        self._rate_limiting_enabled = rate_limit_strategy != "none"
        self._rate_limiting_backoff_rate = env.MERCURY_SYNC_HTTP_RATE_LIMIT_BACKOFF_RATE
        self._max_request_cache_size = env.MERCURY_SYNC_MAX_REQUEST_CACHE_SIZE
        self._request_caching_enabled = env.MERCURY_SYNC_ENABLE_REQUEST_CACHING
        self._verify_cert = env.MERCURY_SYNC_VERIFY_SSL_CERT

    async def connect_async(
        self,
        cert_path: Optional[str] = None,
        key_path: Optional[str] = None,
        worker_socket: Optional[socket.socket] = None,
        upgrade_socket: Optional[socket.socket] = None,
    ):
        async with self._logger.context() as ctx:
            self._backoff_sem = asyncio.Semaphore(self._rate_limiting_backoff_rate)

            await ctx.log(
                Event(
                    level=LogLevel.DEBUG,
                    message=f"Starting HTTP connection on - {self.host}:{self.port}",
                )
            )

            if self._request_caching_enabled:
                self._cache_purge_lock = asyncio.Lock()

            return await super().connect_async(
                cert_path=cert_path,
                key_path=key_path,
                worker_socket=worker_socket,
                upgrade_socket=upgrade_socket
                if upgrade_socket and cert_path and key_path
                else None,
            )

    def read(
        self,
        data: ReceiveBuffer,
        transport: asyncio.Transport,
        data_ready: asyncio.Future,
    ) -> None:
        self._pending_responses.append(
            asyncio.create_task(
                self._route_request(
                    data,
                    transport,
                    data_ready,
                )
            )
        )

    async def _route_request(
        self,
        data: ReceiveBuffer,
        transport: asyncio.Transport,
        data_ready: asyncio.Future,
    ):
        async with self._logger.context() as ctx:
            request_method: bytes | None = None
            request_path: bytes | None = None
            request_version: bytes | None = None
            request_headers: dict[bytes, bytes] = {}
            request_data: bytes | None = None
            request_query: bytes | None = None

            ip_address, _ = transport.get_extra_info("peername")

            try:
                if self._use_encryption:
                    encrypted_data = self._encryptor.decrypt(data)
                    data = self._decompressor.decompress(encrypted_data)

                lines = data.maybe_extract_lines()
                if lines is None and data.is_next_line_obviously_invalid_request_line():
                    raise Exception("Bad request line")

                elif lines is None:
                    raise Exception("No lines received")

                if not lines:
                    raise Exception("Request empty")

                matches = validate(
                    request_line_re, lines[0], "illegal request line: {!r}", lines[0]
                )

                request_headers = {
                    matches["field_name"].decode(): matches["field_value"].decode()
                    for matches in [
                        validate(
                            header_field_re, line, "illegal header line: {!r}", line
                        )
                        for line in _obsolete_line_fold(lines[1:])
                    ]
                }

                if (request_headers.get("Host", "") is None) and (
                    request_headers.get("host") is None
                ):
                    raise Exception("Missing Host header")

                request_method = matches.get("method", b"")
                request_version = matches.get("version", b"HTTP/1.1")
                request_path = matches.get("target", b"")

                request_method = request_method.decode()
                request_path = request_path.decode()

                request_query: Union[str, None] = None
                if "?" in request_path:
                    request_path, request_query = request_path.split("?")

                request_data = b""
                if (
                    (expect_header := request_headers.get("Expect"))
                    and expect_header.lower() == "100-continue"
                    and transport.is_closing() is False
                ):
                    server_error_respnse = HTTPResponse(
                        status=100,
                        protocol="HTTP/1.1",
                        headers=self._response_headers.get(
                            f"{request_method}_{request_path}", {}
                        ),
                        path=request_path,
                        status_message="Continue",
                    )

                    transport.write(server_error_respnse.prepare_response())

                    self.waiting_for_data.set()
                    data = await asyncio.wait_for(
                        data_ready,
                        timeout=self._request_timeout,
                    )

                    request_data += data.read_all()
                    self.waiting_for_data.clear()

                if request_method in [
                    "POST",
                    "PUT",
                    "PATCH",
                    "DELETE",
                ] and (
                    request_headers.get("Transfer-Encoding")
                    or request_headers.get("transfer-encoding")
                ):
                    next_line = bytes(data.maybe_extract_next_line() or b"")
                    chunk_size = int(next_line.rstrip(), 16)

                    while next_line := data.maybe_extract_at_most(chunk_size + 2):
                        next_line = (next_line or b"").rstrip()
                        request_data += next_line

                        chunk_size = int(bytes(data.buffer.rstrip()), 16)

                        if not chunk_size:
                            break

                        self.waiting_for_data.set()
                        data = await asyncio.wait_for(
                            data_ready,
                            timeout=self._request_timeout,
                        )

                        request_data += data.read_all()
                        self.waiting_for_data.clear()

                elif request_method in ["POST", "PUT", "PATCH", "DELETE"] and (
                    (content_length := request_headers.get("Content-Length"))
                    or (content_length := request_headers.get("content-length"))
                ):
                    content_length_amount = int(content_length)

                    request_data += (
                        data.maybe_extract_at_most(content_length_amount + 1) or b""
                    )

                elif request_method in ["POST", "PUT", "PATCH"]:
                    raise Exception(
                        "No Content-Length or Transfer-Encoding header supplied"
                    )

                data.clear()
                await ctx.log(
                    Request(
                        level=LogLevel.DEBUG,
                        message="Request received",
                        path=request_path,
                        method=request_method,
                        ip_address=ip_address,
                    )
                )

            except asyncio.TimeoutError:
                async with self._backoff_sem:
                    data.clear()
                    status = 400

                    self.waiting_for_data.set()

                    await ctx.log(
                        Response(
                            level=LogLevel.ERROR,
                            ip_address=ip_address,
                            error="Timed out.",
                            status=status,
                        ),
                        template="{timestamp} - {level} - {thread_id} - {ip_address}:{status}:{error}",
                    )

                    if transport.is_closing() is False:
                        server_error_respnse = HTTPResponse(
                            status=status,
                            error=f"{status}: Timed out",
                            protocol="HTTP/1.1",
                            headers={"Content-Type": "application/json"},
                        )

                        transport.write(server_error_respnse.prepare_response())

                return

            except Exception as e:
                async with self._backoff_sem:
                    data.clear()
                    status = 400

                    self.waiting_for_data.set()

                    await ctx.log(
                        Response(
                            level=LogLevel.ERROR,
                            ip_address=ip_address,
                            error=str(e),
                            status=status,
                        ),
                        template="{timestamp} - {level} - {thread_id} - {ip_address}:{status}:{error}",
                    )

                    if transport.is_closing() is False:
                        server_error_respnse = HTTPResponse(
                            status=status,
                            error=f"{status}: Bad Request - {str(e)}",
                            protocol="HTTP/1.1",
                        )

                        transport.write(server_error_respnse.prepare_response())

                return

            handler_key = f"{request_method}_{request_path}"

            handler: Handler | None = None
            fabricator: Fabricator | None = None
            request_params: Dict[str, str | Tuple[str]] | None = None

            try:
                handler = self.events[handler_key]
                fabricator = self.fabricators[handler_key]

                await ctx.log(
                    Request(
                        level=LogLevel.DEBUG,
                        message="Exact match for route",
                        method=request_method,
                        path=request_path,
                        ip_address=ip_address,
                    )
                )

            except KeyError:
                # Fallback to Trie router
                if match := self.routes.match(request_path):
                    methods_conifg: Dict[
                        str, Dict[Literal["model", "handler"], Handler | Any]
                    ] = match.anything

                    request_path = match.route
                    handler = methods_conifg.get(request_method)
                    handler_key = f"{request_method}_{request_path}"
                    request_params = match.params
                    fabricator = self.fabricators.get(handler_key)

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message="Fallback match for route",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        )
                    )

            if handler is None:
                async with self._backoff_sem:
                    not_found_response = HTTPResponse(
                        path=request_path,
                        status=404,
                        error="Not Found",
                        protocol=request_version.decode(),
                        method=request_method,
                    )

                    await ctx.log(
                        Response(
                            path=request_path,
                            method=request_method,
                            level=LogLevel.ERROR,
                            ip_address=ip_address,
                            error="Failed to match route",
                            status=404,
                        ),
                        template="{timestamp} - {level} - {thread_id} - {ip_address}:{status} - {method} {path} {error}",
                    )

                    if transport.is_closing() is False:
                        transport.write(not_found_response.prepare_response())

                    return

            elif fabricator is None:
                async with self._backoff_sem:
                    method_not_allowed_response = HTTPResponse(
                        path=request_path,
                        status=405,
                        error="Method Not Allowed",
                        protocol=request_version.decode(),
                        method=request_method,
                    )

                    await ctx.log(
                        Response(
                            path=request_path,
                            method=request_method,
                            level=LogLevel.ERROR,
                            ip_address=ip_address,
                            error="Failed to match allowed methods",
                            status=405,
                        ),
                        template="{timestamp} - {level} - {thread_id} - {ip_address}:{status} - {method} {path} {error}",
                    )

                    if transport.is_closing() is False:
                        transport.write(method_not_allowed_response.prepare_response())

                    return

            try:
                if self._rate_limiting_enabled:
                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message="Entered rate limiting",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                    rejected = await self._limiter.limit(
                        ipaddress.ip_address(ip_address),
                        request_path,
                        request_method,
                        limit=handler.limit,
                    )

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message=f"Rate limiting returned status - {rejected}",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                    if rejected and transport.is_closing() is False:
                        async with self._backoff_sem:
                            await ctx.log(
                                Response(
                                    path=request_path,
                                    method=request_method,
                                    level=LogLevel.ERROR,
                                    ip_address=ip_address,
                                    error="Rejected by rate limiting",
                                    status=429,
                                ),
                                template="{timestamp} - {level} - {thread_id} - {ip_address}:{status} - {method} {path} {error}",
                            )

                            too_many_requests_response = HTTPResponse(
                                path=request_path,
                                status=429,
                                error="Too Many Requests",
                                protocol=request_version.decode(),
                                method=request_method,
                            )

                            transport.write(
                                too_many_requests_response.prepare_response()
                            )

                            return

                    elif rejected:
                        await ctx.log(
                            Response(
                                path=request_path,
                                method=request_method,
                                level=LogLevel.ERROR,
                                ip_address=ip_address,
                                error="Rejected by rate limiting and transport closed - aborting request",
                                status=429,
                            ),
                            template="{timestamp} - {level} - {thread_id} - {ip_address}:{status} - {method} {path} {error}",
                        )

                        async with self._backoff_sem:
                            transport.close()

                            return

                has_middleware = self._middleware_enabled.get(handler_key)

                await ctx.log(
                    Request(
                        level=LogLevel.DEBUG,
                        message=(
                            "Middleware found"
                            if has_middleware
                            else "No middleware found"
                        ),
                        method=request_method,
                        path=request_path,
                        ip_address=ip_address,
                    ),
                )

                args, kwargs, validation_error = fabricator.parse(
                    request_data,
                    request_headers=request_headers,
                    request_query=request_query,
                    request_params=request_params,
                    has_middleware=has_middleware,
                )

                await ctx.log(
                    Request(
                        level=LogLevel.DEBUG,
                        message="Fabricator parsed args",
                        method=request_method,
                        path=request_path,
                        ip_address=ip_address,
                    ),
                )

                if validation_error:
                    await ctx.log(
                        Response(
                            path=request_path,
                            method=request_method,
                            level=LogLevel.ERROR,
                            ip_address=ip_address,
                            error=f"Rejected by request validation - {str(validation_error)}",
                            status=422,
                        ),
                        template="{timestamp} - {level} - {ip_address}:{status} - {method} {path} {error}",
                    )

                    invalid_request_response = HTTPResponse(
                        path=request_path,
                        status=422,
                        error=f"Validation error - {str(validation_error)}",
                        protocol=request_version.decode(),
                        method=request_method,
                        headers={"content-type": "application/json"},
                    )

                    transport.write(invalid_request_response.prepare_response())

                    return

                response_headers: Dict[str, str] = self._response_headers.get(
                    handler_key, {}
                )
                encoded_data: str = ""

                await ctx.log(
                    Request(
                        level=LogLevel.DEBUG,
                        message=(
                            f"Request - {request_method} {request_path}:{ip_address} - sending response headers - {', '.join(response_headers)}"
                            if len(response_headers) > 0
                            else f"Request - {request_method} {request_path}:{ip_address} - has no response headers"
                        ),
                        method=request_method,
                        path=request_path,
                        ip_address=ip_address,
                    ),
                )

                (response_parser, status_code) = self._response_parsers.get(
                    handler_key, (None, None)
                )

                if status_code is None:
                    status_code = 200
                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message=f"Set status code to default - {status_code}",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                transport.get_extra_info("")

                if has_middleware:
                    context = ResponseContext(
                        request_path,
                        request_method,
                        request_headers,
                        request_params,
                        request_query,
                        request_data,
                        args,
                        kwargs,
                        fabricator,
                        response_parser,
                        ip_address,
                        "https"
                        if bool(transport.get_extra_info("sslcontext"))
                        else "http",
                        transport.get_extra_info("sockname"),
                        self._upgrade_port,
                    )

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message=f"Executing route handler with middleware - {handler.__class__.__name__}",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                    response: Tuple[ResponseContext, Any] = await handler(
                        context=context
                    )

                    context, response_data = response

                    if len(context.errors) > 0 and transport.is_closing() is False:
                        error_headers = {"content-type": "application/json"}

                        errors = orjson.dumps(
                            [
                                {
                                    "error": str(error),
                                }
                                for error in context.errors
                            ]
                        )

                        middleware_error_response = HTTPResponse(
                            path=request_path,
                            status=500,
                            error="Internal server error.",
                            data=errors.decode(),
                            protocol=request_version.decode(),
                            headers=error_headers,
                            method=request_method,
                        )

                        joined_errors = ", ".join(
                            [str(error) for error in context.errors]
                        )
                        await ctx.log(
                            Response(
                                path=request_path,
                                method=request_method,
                                level=LogLevel.ERROR,
                                ip_address=ip_address,
                                error=f"Middleware encountered errors - {joined_errors}",
                                status=500,
                            ),
                            template="{timestamp} - {level} - {thread_id} - {ip_address}:{status} - {method} {path} {error}",
                        )

                        transport.write(
                            middleware_error_response.prepare_response(
                                compression=context.compressor,
                                compression_level=context.compression_level,
                            )
                        )

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message="Completed route handler execution with middleware",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                    response_headers.update(context.response_headers)
                    status_code = context.status or status_code

                else:
                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message="Executing route handler",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                    response_data = await handler(*args, **kwargs)

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message="Completed route handler execution",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                if response_parser:
                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message="Serializing response body",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                    encoded_data = parse_response(response_data, response_parser)
                    content_length = len(encoded_data)
                    headers = f"content-length: {content_length}"

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message=f"Serialized {content_length} bytes",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                elif response_data:
                    encoded_data = response_data

                    content_length = len(response_data)

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message=f"Set response body as {content_length} bytes",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                    headers = f"content-length: {content_length}"

                else:
                    headers = "content-length: 0"

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message="Response has no body",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                for key in response_headers:
                    headers = f"{headers}\r\n{key}: {response_headers[key]}"

                    await ctx.log(
                        Request(
                            level=LogLevel.DEBUG,
                            message=f"Response adding header - {key}:{response_headers[key]}",
                            method=request_method,
                            path=request_path,
                            ip_address=ip_address,
                        ),
                    )

                response_data = f"HTTP/1.1 {status_code} OK\r\n{headers}\r\n\r\n{encoded_data}".encode()

                if self._use_encryption:
                    encrypted_data = self._encryptor.encrypt(response_data)
                    response_data = self._compressor.compress(encrypted_data)

                if self._use_encryption is False:
                    await ctx.log(
                        Response(
                            path=request_path,
                            method=request_method,
                            ip_address=ip_address,
                            status=status_code,
                        ),
                        template="{timestamp} - {level} - {thread_id} - {ip_address}:{status} - {method} {path} - {status}",
                    )

                transport.write(response_data)

            except Exception as e:
                async with self._backoff_sem:
                    await ctx.log(
                        Response(
                            path=request_path,
                            method=request_method,
                            level=LogLevel.ERROR,
                            ip_address=ip_address,
                            error=str(e),
                            status=500,
                        ),
                        template="{timestamp} - {level} - {thread_id} - {ip_address}:{status} - {method} {path} {error}",
                    )

                    if transport.is_closing() is False:
                        server_error_respnse = HTTPResponse(
                            path=request_path,
                            status=500,
                            error=f"Internal Error - {str(e)}",
                            protocol=request_version.decode(),
                            headers={"content-type": "application/json"},
                            method=request_method,
                        )

                        transport.write(server_error_respnse.prepare_response())

    async def close(self):
        await self._limiter.close()
        await super().close()

        await self._logger.log(
            Event(
                level=LogLevel.DEBUG,
                message=f"Closing HTTP connection at - {self.host}:{self.port}",
            )
        )

        await self._logger.close()

    def abort(self):
        self._limiter.abort()
        super().abort()

        self._logger.abort()



from __future__ import annotations
import msgspec
from http.cookies import SimpleCookie
from typing import (
    Any,
    Dict,
    List,
    Type,
    TypeVar,
)

import orjson
from pydantic import (
    StrictBool,
    StrictFloat,
    StrictInt,
    StrictStr,
)
from .model import Model

HTTPEncodable = StrictStr | StrictInt | StrictBool | StrictFloat | None

T = TypeVar("t")


class Headers(Model):
    @classmethod
    def make(cls, model: type[Headers], headers: dict[str, str]):
        return model(**headers)

    @classmethod
    def make_raw(cls, headers: dict[str, str]):
        return headers


class Cookies(Model):
    @classmethod
    def make(cls, model: Type[Cookies], headers: Dict[str, str]):
        cookie_header = headers.get("cookie")

        if cookie_header is None:
            return model()

        parsed_cookies = SimpleCookie()
        parsed_cookies.load(cookie_header)

        return model(**{name: morsel.value for name, morsel in parsed_cookies.items()})

    @classmethod
    def make_raw(cls, headers: Dict[str, Any]):
        cookie_header: str = headers.get("cookie")

        if cookie_header is None:
            return

        parsed_cookies = SimpleCookie()
        parsed_cookies.load(cookie_header)

        return {name: morsel.value for name, morsel in parsed_cookies.items()}


class Parameters(Model):
    @classmethod
    def make(cls, model: type[Parameters], params: dict[str, str]):
        return model(**params)

    @classmethod
    def make_raw(cls, params: dict[str, str]):
        return params


class Query(Model):
    @classmethod
    def make(cls, model: Type[Query], query: str):
        query_params: Dict[str, str] = {}
        if len(query) < 1:
            params = query.split("&")

            for param in params:
                key, value = param.split("=")

                query_params[key] = value

        return model(**query_params)

    @classmethod
    def make_raw(cls, query: str):
        query_params: Dict[str, str] = {}
        if len(query) < 1:
            params = query.split("&")

            for param in params:
                key, value = param.split("=")

                query_params[key] = value

        return query_params


class Body(Model):
    content: bytes

    def text(self, encoding: str = "utf-8"):
        return self.content.decode(encoding=encoding)

    def json(self):
        return orjson.loads(self.content)

    @classmethod
    def make(
        cls,
        data: bytes,
    ):
        return Body(content=data)

    @classmethod
    def make_raw(cls, data: bytes):
        return data



from .file_upload import FileUpload as FileUpload
from .html import HTML as HTML
from .http_response import HTTPResponse as HTTPResponse
from .limit import Limit as Limit
from .parse_response import parse_response as parse_response
from .request_models import Body as Body
from .request_models import Cookies as Cookies
from .request_models import Headers as Headers
from .request_models import Parameters as Parameters
from .request_models import Query as Query
from .model import Model as Model



from typing import List

from pydantic import BaseModel

from .models import Schema
from .parsed_types import (
    FieldMetadata,
    FieldType,
)


def parse_type(field: FieldMetadata) -> List[str] | str:
    field_data: List[FieldType] = field.get("anyOf")
    if field_data:
        return {"enum": [field_type.get("type") for field_type in field_data]}

    return {"type": field.get("type", "string")}


def parse_model(model: type[BaseModel]):
    return Schema(**model.model_json_schema())



from __future__ import annotations
from typing import Any, Dict, Literal

from mkfst.logging import Entry, LogLevel

JSONEncodableValue = str | int | bool | float | None


class Event(Entry, kw_only=True):
    message: str | bytes | None = None
    level: LogLevel = LogLevel.INFO

    def to_template(self, template: str, context=None):
        kwargs: dict[
            str,
            int | str | bool | float | LogLevel | list | dict | set | Any,
        ] = {field: getattr(self, field) for field in self.__struct_fields__}

        if isinstance(self.message, bytes):
            kwargs["message"] = self.message.decode()

        kwargs["level"] = kwargs["level"].value

        if context:
            kwargs.update(context)

        return template.format(**kwargs)


class Request(Entry, kw_only=True):
    level: LogLevel = LogLevel.INFO
    message: str | bytes | None = None
    path: str | bytes | None = None
    protocol: Literal["HTTP/1.1"] | Literal[b"HTTP/1.1"] = "HTTP/1.1"
    method: str | bytes | None = None
    headers: Dict[str, JSONEncodableValue] | dict[bytes, bytes] | None = None
    params: Dict[str, JSONEncodableValue] | dict[bytes, bytes] | None = None
    query: str | bytes | None = None
    body: str | bytes | None = None
    error: Exception | str | None = None
    ip_address: str | None = None

    def to_template(
        self,
        template: str,
        context: Dict[str, Any] | None = None,
    ):
        kwargs = {
            "level": self.level.value,
            "path": self.path
            if isinstance(self.path, str) or self.path is None
            else self.path.decode(),
            "method": self.method
            if isinstance(self.method, str) or self.method is None
            else self.method.decode(),
            "headers": ", ".join(
                [
                    f"{key}:{header}"
                    if isinstance(key, str) and isinstance(header, str)
                    else f"{key.decode()}:{header.decode()}"
                    for key, header in self.headers.items()
                ]
            )
            if self.headers
            else None,
            "params": ", ".join(
                [
                    f"{key}:{param}"
                    if isinstance(key, str) and isinstance(param, str)
                    else f"{key.decode()}:{param.decode()}"
                    for key, param in self.params.items()
                ]
            )
            if self.params
            else None,
            "query": self.query
            if isinstance(self.query, str) or self.query is None
            else self.query.decode(),
            "body": self.body,
            "error": self.error,
            "ip_address": self.ip_address,
        }

        if isinstance(self.message, bytes):
            kwargs["message"] = self.message.decode()

        elif self.message:
            kwargs["message"] = self.message

        if context:
            kwargs.update(context)

        return template.format(**kwargs)


class Response(Entry, kw_only=True):
    level: LogLevel = LogLevel.INFO
    message: str | bytes | None = None
    path: str | bytes | None = None
    protocol: Literal["HTTP/1.1"] | Literal[b"HTTP/1.1"] = "HTTP/1.1"
    method: str | bytes | None = None
    error: Exception | str | None = None
    ip_address: str | None = None
    status: int | None = None

    def to_template(
        self,
        template: str,
        context: Dict[str, Any] | None = None,
    ):
        kwargs: Dict[
            str,
            int | str | bool | float | LogLevel | list | dict | set | Any,
        ] = {
            field: getattr(self, field)
            for field in self.__struct_fields__
            if getattr(self, field) is not None
        }

        if isinstance(self.path, bytes):
            kwargs["path"] = self.path.decode()

        if isinstance(self.protocol, bytes):
            kwargs["protocol"] = self.protocol.decode()

        if isinstance(self.method, bytes):
            kwargs["method"] = self.method.decode()

        if isinstance(self.message, bytes):
            kwargs["message"] = self.message.decode()

        kwargs["level"] = kwargs["level"].value

        if context:
            kwargs.update(context)

        return template.format(**kwargs)



from __future__ import annotations
import re

blank_line_regex = re.compile(b"\n\r?\n", re.MULTILINE)


class ReceiveBuffer:
    def __init__(self) -> None:
        self.buffer = bytearray()
        self._next_line_search = 0
        self._multiple_lines_search = 0

    def __iadd__(self, byteslike: bytes | bytearray) -> "ReceiveBuffer":
        self.buffer += byteslike
        return self

    def __bool__(self) -> bool:
        return bool(len(self))

    def __len__(self) -> int:
        return len(self.buffer)

    # for @property unprocessed_data
    def __bytes__(self) -> bytes:
        return bytes(self.buffer)

    def _extract(self, count: int) -> bytearray:
        # extracting an initial slice of the data buffer and return it
        out = self.buffer[:count]
        del self.buffer[:count]

        self._next_line_search = 0
        self._multiple_lines_search = 0

        return out

    def read_all(self) -> bytearray:
        data = bytes(self.buffer)
        self.buffer.clear()

        self._next_line_search = 0
        self._multiple_lines_search = 0

        return data

    def maybe_extract_at_most(self, count: int) -> bytearray | None:
        """
        Extract a fixed number of bytes from the buffer.
        """
        out = self.buffer[:count]
        if not out:
            return None

        out = self.buffer[:count]
        del self.buffer[:count]

        self._next_line_search = 0
        self._multiple_lines_search = 0

        return out

    def maybe_extract_next_line(self) -> bytearray | None:
        """
        Extract the first line, if it is completed in the buffer.
        """
        # Only search in buffer space that we've not already looked at.
        search_start_index = max(0, self._next_line_search - 1)
        partial_idx = self.buffer.find(b"\r\n", search_start_index)

        if partial_idx == -1:
            self._next_line_search = len(self.buffer)
            return None

        # + 2 is to compensate len(b"\r\n")
        idx = partial_idx + 2

        out = self.buffer[:idx]
        del self.buffer[:idx]

        self._next_line_search = 0
        self._multiple_lines_search = 0

        return out

    def maybe_extract_lines(self) -> list[bytearray] | None:
        """
        Extract everything up to the first blank line, and return a list of lines.
        """
        # Handle the case where we have an immediate empty line.
        if self.buffer[:1] == b"\n":
            out = self.buffer[:1]
            del self.buffer[:1]

            self._next_line_search = 0
            self._multiple_lines_search = 0

            return []

        if self.buffer[:2] == b"\r\n":
            out = self.buffer[:2]
            del self.buffer[:2]

            self._next_line_search = 0
            self._multiple_lines_search = 0

            return []

        # Only search in buffer space that we've not already looked at.
        match = blank_line_regex.search(self.buffer, self._multiple_lines_search)
        if match is None:
            self._multiple_lines_search = max(0, len(self.buffer) - 2)
            return None

        # Truncate the buffer and return it.
        idx = match.span(0)[-1]
        out = self.buffer[:idx]
        del self.buffer[:idx]

        self._next_line_search = 0
        self._multiple_lines_search = 0

        lines = out.split(b"\n")

        for line in lines:
            if line.endswith(b"\r"):
                del line[-1]

        assert lines[-2] == lines[-1] == b""

        del lines[-2:]

        return lines

    # In theory we should wait until `\r\n` before starting to validate
    # incoming data. However it's interesting to detect (very) invalid data
    # early given they might not even contain `\r\n` at all (hence only
    # timeout will get rid of them).
    # This is not a 100% effective detection but more of a cheap sanity check
    # allowing for early abort in some useful cases.
    # This is especially interesting when peer is messing up with HTTPS and
    # sent us a TLS stream where we were expecting plain HTTP given all
    # versions of TLS so far start handshake with a 0x16 message type code.
    def is_next_line_obviously_invalid_request_line(self) -> bool:
        try:
            # HTTP header line must not contain non-printable characters
            # and should not start with a space
            return self.buffer[0] < 0x21
        except IndexError:
            return False

    def clear(self):
        self.buffer.clear()
        self._next_line_search = 0
        self._multiple_lines_search = 0



from __future__ import annotations
import msgspec
from base64 import b64encode
from gzip import compress as gzip_compress
from typing import Dict, Literal, Optional

import orjson
from zstandard import compress as zstd_compress

from .model import Model


class HTTPResponse(msgspec.Struct, kw_only=True):
    protocol: str = "HTTP/1.1"
    path: Optional[str] = None
    error: Optional[str] = None
    method: Optional[
        Literal["GET", "POST", "HEAD", "OPTIONS", "PUT", "PATCH", "DELETE"]
    ] = None
    status: Optional[int] = None
    status_message: Optional[str] = None
    params: Dict[str, str] = {}
    headers: Dict[str, str] = {}
    data: Optional[str | Model] = None

    def prepare_response(
        self,
        compression: Literal["gzip", "zstd"] | None = None,
        compression_level: int | None = None,
    ):
        message = "OK"
        if self.error:
            message = self.error

        head_line = f"HTTP/1.1 {self.status} {message}"

        encoded_data: str = ""

        if isinstance(self.data, Model) or self.data in Model.__subclasses__():
            encoded_data = orjson.dumps(msgspec.structs.asdict(self.data)).decode()
            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        elif isinstance(self.data, (dict, list)):
            encoded_data = orjson.dumps(self.data).decode()
            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        elif self.data:
            encoded_data = self.data

            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        else:
            headers = "content-length: 0"

        if compression == "gzip":
            encoded_data = b64encode(
                gzip_compress(encoded_data.encode(), compresslevel=compression_level)
            ).decode()

            headers = f"{headers}\r\nx-compression-encoding: {compression}"
            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        elif compression == "zstd":
            encoded_data = b64encode(
                zstd_compress(encoded_data.encode(), level=compression_level)
            ).decode()

            headers = f"{headers}\r\nx-compression-encoding: {compression}"
            content_length = len(encoded_data)
            headers = f"content-length: {content_length}"

        response_headers = self.headers
        if response_headers:
            for key in response_headers:
                headers = f"{headers}\r\n{key}: {response_headers[key]}"

        return f"{head_line}\r\n{headers}\r\n\r\n{encoded_data}".encode()



from typing import Any, Dict, Type
import msgspec
import orjson

from .file_upload import FileUpload
from .html import HTML
from .request_models import Body
from .model import Model


def parse_response(
    response: Model | Dict[Any, Any] | str,
    response_model: Type[Model | Body | FileUpload | HTML | dict | list | str | bytes],
) -> str:
    if response_model == HTML or isinstance(response, HTML):
        return response.format()

    elif (
        response_model == FileUpload or response_model in FileUpload.__subclasses__()
    ) or isinstance(response, FileUpload):
        return response.data.decode(response.encoding)

    elif (
        response_model == Body
        or response_model in Body.__subclasses__()
        or isinstance(response, Body)
    ):
        return response.content.decode()

    elif response_model == dict or response_model == list:
        return orjson.dumps(response).decode()

    elif response_model in Model.__subclasses__():
        return orjson.dumps(response.model_dump()).decode()

    return response



from typing import Callable, Dict, Optional, Union

from pydantic import IPvAnyAddress

from mkfst.env import Env
from mkfst.logging import Logger, LogLevel
from mkfst.models.http import Limit
from mkfst.models.logging import Event

from .limiters import (
    AdaptiveRateLimiter,
    CPUAdaptiveLimiter,
    LeakyBucketLimiter,
    ResourceAdaptiveLimiter,
    SlidingWindowLimiter,
    TokenBucketLimiter,
)


class Limiter:
    def __init__(self, env: Env) -> None:
        self._limiter: Union[
            Union[
                AdaptiveRateLimiter,
                CPUAdaptiveLimiter,
                LeakyBucketLimiter,
                ResourceAdaptiveLimiter,
                SlidingWindowLimiter,
                TokenBucketLimiter,
            ],
            None,
        ] = None

        self._default_limit = Limit(
            max_requests=env.MERCURY_SYNC_HTTP_RATE_LIMIT_REQUESTS,
            request_period=env.MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD,
            reject_requests=env.MERCURY_SYNC_HTTP_RATE_LIMIT_DEFAULT_REJECT,
            cpu_limit=env.MERCURY_SYNC_HTTP_CPU_LIMIT,
            memory_limit=env.MERCURY_SYNC_HTTP_MEMORY_LIMIT,
        )

        self._rate_limit_strategy = env.MERCURY_SYNC_HTTP_RATE_LIMIT_STRATEGY
        self._default_limiter_type = env.MERCURY_SYNC_HTTP_RATE_LIMITER_TYPE

        self._rate_limiter_types: Dict[
            str,
            Callable[
                [Limit],
                Union[
                    AdaptiveRateLimiter,
                    CPUAdaptiveLimiter,
                    LeakyBucketLimiter,
                    ResourceAdaptiveLimiter,
                    SlidingWindowLimiter,
                    TokenBucketLimiter,
                ],
            ],
        ] = {
            "adaptive": AdaptiveRateLimiter,
            "cpu-adaptive": CPUAdaptiveLimiter,
            "leaky-bucket": LeakyBucketLimiter,
            "rate-adaptive": ResourceAdaptiveLimiter,
            "sliding-window": SlidingWindowLimiter,
            "token-bucket": TokenBucketLimiter,
        }

        self._rate_limit_period = env.MERCURY_SYNC_HTTP_RATE_LIMIT_PERIOD

        self._rate_limiters: Dict[
            str,
            Union[
                AdaptiveRateLimiter,
                CPUAdaptiveLimiter,
                LeakyBucketLimiter,
                SlidingWindowLimiter,
                TokenBucketLimiter,
            ],
        ] = {}

        self._logger = Logger()

    async def limit(
        self, 
        ip_address: IPvAnyAddress, 
        path: str, 
        method: str, 
        limit: Optional[Limit] = None
    ):
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            
            rejected = False
            limit_type = limit.limiter_type if limit else self._default_limiter_type

            if limit is None:
                limit = self._default_limit
            
            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {method} {path}:{str(ip_address)} - Using {limit_type} rate limiter with strategy - {self._rate_limit_strategy}'
            ))
            
            limit_key: Union[str, None] = None

            if self._rate_limit_strategy == "ip":
                limit_key = limit.get_key(
                    path,
                    method, 
                    ip_address, 
                    default=ip_address
                )

            elif self._rate_limit_strategy == "endpoint":
                limit_key = limit.get_key(
                    path,
                    method, 
                    ip_address, 
                    default=path
                )

            elif self._rate_limit_strategy == "global":
                limit_key = self._default_limit.get_key(
                    path, 
                    method, 
                    ip_address, 
                    default="default",
                )

                limit = self._default_limit

            elif self._rate_limit_strategy == "ip-endpoint":
                limit_key = limit.get_key(
                    path,
                    method, 
                    ip_address, 
                    default=f"{path}_{ip_address}"
                )

            elif limit:
                limit_key = limit.get_key(
                    path, 
                    method, 
                    ip_address,
                )

            if limit_key and limit.matches(
                path,
                method, 
                ip_address,
            ):
                rejected = await self._check_limiter(
                    limit_key, 
                    limit,
                    path,
                    method,
                    ip_address,
                )

            if rejected:
                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {method} {path}:{str(ip_address)} - Request rejected by {limit.limiter_type} using strategy {self._rate_limit_strategy}'
                ))

            else:
                await ctx.log(Event(
                    level=LogLevel.DEBUG,
                    message=f'Request - {method} {path}:{str(ip_address)} - Request accepted by {limit.limiter_type} using strategy {self._rate_limit_strategy}'
                ))

            return rejected

    async def _check_limiter(
        self, 
        limiter_key: str, 
        limit: Limit,
        path: str,
        method: str,
        ip_address: IPvAnyAddress
    ):
        async with self._logger.context(
            template="{timestamp} - {level} - {thread_id} - {message}",
        ) as ctx:
            
            await ctx.log(Event(
                level=LogLevel.DEBUG,
                message=f'Request - {method} {path}:{str(ip_address)} - Checking  {limit.limiter_type} rate limiter with strategy - {self._rate_limit_strategy}'
            ))
            
            limiter = self._rate_limiters.get(limiter_key)

            rate_limiter_type = limit.limiter_type
            if rate_limiter_type is None:
                rate_limiter_type = self._default_limiter_type

            if limiter is None:
                limiter = self._rate_limiter_types.get(rate_limiter_type)(limit)

                self._rate_limiters[limiter_key] = limiter

            return await limiter.acquire()

    async def close(self):
        for limiter in self._rate_limiters.values():
            if isinstance(limiter, CPUAdaptiveLimiter):
                await limiter.close()

    def abort(self):
        for limiter in self._rate_limiters.values():
            if isinstance(limiter, CPUAdaptiveLimiter):
                limiter.abort()
